from asyncio import Lock, gather
from collections.abc import Callable
from pathlib import Path
from typing import Protocol, overload, runtime_checkable
from uuid import UUID, uuid4

from draive.evaluation.scenario import ScenarioEvaluatorResult
from draive.parameters import DataModel, Field
from draive.scope import ctx
from draive.types import frozenlist
from draive.utils import asynchronous

__all__ = [
    "evaluation_suite",
    "EvaluationCaseResult",
    "EvaluationSuite",
    "EvaluationSuiteCaseResult",
    "EvaluationSuiteDefinition",
    "EvaluationSuiteStorage",
]


class EvaluationSuiteCase[CaseParameters: DataModel](DataModel):
    identifier: UUID = Field(default_factory=uuid4)
    parameters: CaseParameters
    comment: str | None = None


class EvaluationSuiteCaseResult[CaseParameters: DataModel, Value: DataModel | str](DataModel):
    case: EvaluationSuiteCase[CaseParameters] = Field(
        description="Evaluated case",
    )
    value: Value = Field(
        description="Evaluated value",
    )
    results: frozenlist[ScenarioEvaluatorResult] = Field(
        description="Evaluation results",
    )

    @property
    def passed(self) -> bool:
        return all(result.passed for result in self.results)


class EvaluationCaseResult[Value: DataModel | str](DataModel):
    value: Value = Field(
        description="Evaluated value",
    )
    results: frozenlist[ScenarioEvaluatorResult] = Field(
        description="Evaluation results",
    )


@runtime_checkable
class EvaluationSuiteDefinition[CaseParameters: DataModel, Value: DataModel | str](Protocol):
    async def __call__(
        self,
        evaluation_case: CaseParameters,
    ) -> EvaluationCaseResult[Value]: ...


class EvaluationSuiteData[CaseParameters: DataModel](DataModel):
    cases: frozenlist[EvaluationSuiteCase[CaseParameters]] = Field(default_factory=tuple)


@runtime_checkable
class EvaluationSuiteStorage[CaseParameters: DataModel](Protocol):
    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]: ...

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None: ...


class EvaluationSuite[CaseParameters: DataModel, Value: DataModel | str]:
    def __init__(
        self,
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
        storage: EvaluationSuiteStorage[CaseParameters],
    ) -> None:
        self._definition: EvaluationSuiteDefinition[CaseParameters, Value] = definition
        self._storage: EvaluationSuiteStorage[CaseParameters] = storage
        self._data_cache: EvaluationSuiteData[CaseParameters] | None = None
        self._lock: Lock = Lock()

    @overload
    async def __call__(
        self,
        *,
        evaluated_case: CaseParameters | UUID | None,
        reload: bool = False,
    ) -> EvaluationSuiteCaseResult[CaseParameters, Value]: ...

    @overload
    async def __call__(
        self,
        *,
        reload: bool = False,
    ) -> list[EvaluationSuiteCaseResult[CaseParameters, Value]]: ...

    async def __call__(
        self,
        *,
        evaluated_case: CaseParameters | UUID | None = None,
        reload: bool = False,
    ) -> (
        list[EvaluationSuiteCaseResult[CaseParameters, Value]]
        | EvaluationSuiteCaseResult[CaseParameters, Value]
    ):
        async with self._lock:
            match evaluated_case:
                case None:
                    return await gather(
                        *[
                            self._evaluate(evaluated_case=case)
                            for case in (await self._data(reload=reload)).cases
                        ],
                        return_exceptions=False,
                    )

                case UUID() as identifier:
                    available_cases: frozenlist[EvaluationSuiteCase[CaseParameters]] = (
                        await self._data(reload=reload)
                    ).cases

                    if evaluation_case := next(
                        iter([case for case in available_cases if case.identifier == identifier]),
                        None,
                    ):
                        return await self._evaluate(evaluated_case=evaluation_case)

                    else:
                        raise ValueError(f"Evaluation case with ID {identifier} does not exists.")

                case case_parameters:
                    return await self._evaluate(
                        evaluated_case=EvaluationSuiteCase[CaseParameters](
                            parameters=case_parameters,
                        )
                    )

    async def _evaluate(
        self,
        *,
        evaluated_case: EvaluationSuiteCase[CaseParameters],
    ) -> EvaluationSuiteCaseResult[CaseParameters, Value]:
        case_result: EvaluationCaseResult[Value] = await self._definition(
            evaluation_case=evaluated_case.parameters
        )

        return EvaluationSuiteCaseResult[CaseParameters, Value](
            case=evaluated_case,
            value=case_result.value,
            results=case_result.results,
        )

    async def _data(
        self,
        reload: bool = False,
    ) -> EvaluationSuiteData[CaseParameters]:
        if (data := self._data_cache) and not reload:
            return data

        else:
            self._data_cache = await self._storage.load()
            return self._data_cache

    async def cases(
        self,
        reload: bool = False,
    ) -> frozenlist[EvaluationSuiteCase[CaseParameters]]:
        async with self._lock:
            return (await self._data(reload=reload)).cases

    async def add_case(
        self,
        parameters: CaseParameters,
        /,
        comment: str | None = None,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=(
                    *data.cases,
                    EvaluationSuiteCase[CaseParameters](
                        parameters=parameters,
                        comment=comment,
                    ),
                )
            )
            await self._storage.save(self._data_cache)

    async def remove_case(
        self,
        identifier: UUID,
        /,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=tuple(case for case in data.cases if case.identifier != identifier)
            )
            await self._storage.save(self._data_cache)


def evaluation_suite[CaseParameters: DataModel, Value: DataModel | str](
    case: type[CaseParameters],
    /,
    storage: EvaluationSuiteStorage[CaseParameters] | Path | str | None = None,
) -> Callable[
    [EvaluationSuiteDefinition[CaseParameters, Value]],
    EvaluationSuite[CaseParameters, Value],
]:
    suite_storage: EvaluationSuiteStorage[CaseParameters]
    match storage:
        case None:
            suite_storage = _EvaluationSuiteMemoryStorage[CaseParameters](
                data_type=EvaluationSuiteData[case],
            )

        case str() as path_str:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path_str,
                data_type=EvaluationSuiteData[case],
            )

        case Path() as path:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path,
                data_type=EvaluationSuiteData[case],
            )

        case storage:
            suite_storage = storage

    def wrap(
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
    ) -> EvaluationSuite[CaseParameters, Value]:
        return EvaluationSuite[CaseParameters, Value](
            definition=definition,
            storage=suite_storage,
        )

    return wrap


class _EvaluationSuiteMemoryStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self.store: frozenlist[EvaluationSuiteCase[CaseParameters]] = ()
        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        return self._data_type(cases=self.store)

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        self.store = data.cases


class _EvaluationSuiteFileStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        path: Path | str,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        try:
            return await self._file_load(data_type=self._data_type)

        except ValueError as exc:
            ctx.log_error(
                f"Invalid EvaluationSuite at {self._path}",
                exception=exc,
            )
            return self._data_type()

    @asynchronous(executor=None)
    def _file_load(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> EvaluationSuiteData[CaseParameters]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                return data_type.from_json(file.read())

        else:
            return data_type()

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        await self._file_save(data=data)

    @asynchronous(executor=None)
    def _file_save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        with open(self._path, mode="wb+") as file:
            file.write(data.as_json(indent=2).encode("utf-8"))



from pathlib import Path
from typing import Any

from draive.mistral.config import MistralChatConfig
from draive.scope import ctx
from draive.sentencepiece import sentencepiece_tokenize_text
from draive.utils import cache

__all__ = [
    "mistral_tokenize_text",
]


def mistral_tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    return sentencepiece_tokenize_text(
        text=text,
        model_path=_model_path(ctx.state(MistralChatConfig).model),
        **extra,
    )


@cache(limit=4)
def _model_path(model_name: str) -> str:
    model_file: str = _mapping.get(model_name, "mistral_v3_tokenizer.model")
    return str(Path(__file__).parent / "tokens" / model_file)


_mapping: dict[str, str] = {
    "open-mistral-7b": "mistral_v1_tokenizer.model",
    "open-mixtral-8x7b": "mistral_v1_tokenizer.model",
    "mistral-embed": "mistral_v1_tokenizer.model",
    "mistral-small": "mistral_v2_tokenizer.model",
    "mistral-large": "mistral_v2_tokenizer.model",
    "open-mixtral-8x22b": "mistral_v3_tokenizer.model",
    "codestral-22b": "mistral_v3_tokenizer.model",
}



from collections.abc import Iterable, Iterator
from itertools import chain
from typing import Self, final, overload

from draive.parameters.model import DataModel
from draive.types.audio import AudioBase64Content, AudioContent, AudioURLContent
from draive.types.frozenlist import frozenlist
from draive.types.image import ImageBase64Content, ImageContent, ImageURLContent
from draive.types.text import TextContent
from draive.types.video import VideoBase64Content, VideoContent, VideoURLContent

__all__ = [
    "MultimodalContent",
    "MultimodalContentElement",
    "MultimodalContentConvertible",
]

MultimodalContentElement = TextContent | ImageContent | AudioContent | VideoContent | DataModel
MultimodalContentConvertible = str | MultimodalContentElement


@final
class MultimodalContent(DataModel):
    @classmethod
    def of(
        cls,
        *elements: Self | MultimodalContentConvertible,
        merge_text: bool = False,
    ) -> Self:
        match elements:
            case [MultimodalContent() as content] if not merge_text:
                return content  # pyright: ignore[reportReturnType]

            case elements:
                if merge_text:
                    return cls(
                        parts=tuple(
                            _merge_texts(
                                *chain.from_iterable(
                                    _extract_parts(element) for element in elements
                                )
                            )
                        ),
                    )

                else:
                    return cls(
                        parts=tuple(
                            chain.from_iterable(_extract_parts(element) for element in elements)
                        ),
                    )

    parts: frozenlist[MultimodalContentElement]

    @property
    def has_media(self) -> bool:
        return any(_is_media(part) for part in self.parts)

    @property
    def has_artifacts(self) -> bool:
        return any(_is_artifact(part) for part in self.parts)

    @overload
    def artifacts(self, /) -> frozenlist[DataModel]: ...

    @overload
    def artifacts[Artifact: DataModel](
        self,
        model: type[Artifact],
        /,
    ) -> frozenlist[Artifact]: ...

    def artifacts[Artifact: DataModel](
        self,
        model: type[Artifact] | None = None,
        /,
    ) -> frozenlist[Artifact] | frozenlist[DataModel]:
        if model:
            return tuple(part for part in self.parts if isinstance(part, model))

        else:
            return tuple(part for part in self.parts if _is_artifact(part))

    def excluding_artifacts(self) -> Self:
        return self.__class__(
            parts=tuple(part for part in self.parts if not _is_artifact(part)),
        )

    def as_string(
        self,
        joiner: str | None = None,
    ) -> str:
        return (joiner or "\n").join(_as_string(element) for element in self.parts)

    def appending(
        self,
        *parts: MultimodalContentConvertible,
        merge_text: bool = False,
    ) -> Self:
        if not self.parts:
            if merge_text:
                return self.__class__(
                    parts=tuple(_merge_texts(*(_as_content(element) for element in parts))),
                )

            else:
                return self.__class__(
                    parts=tuple(_as_content(element) for element in parts),
                )

        if merge_text:
            # check the last part
            match self.parts[-1]:
                case TextContent() as text:
                    # if it is a text append merge starting with it
                    return self.__class__(
                        parts=(
                            *self.parts[:-1],
                            *_merge_texts(text, *(_as_content(element) for element in parts)),
                        )
                    )

                case _:
                    # otherwise just append merged items
                    return self.__class__(
                        parts=(
                            *self.parts,
                            *_merge_texts(*(_as_content(element) for element in parts)),
                        )
                    )

        else:
            return self.__class__(
                parts=(
                    *self.parts,
                    *(_as_content(element) for element in parts),
                )
            )

    def extending(
        self,
        *other: Self,
        merge_text: bool = False,
    ) -> Self:
        return self.appending(
            *chain.from_iterable(content.parts for content in other),
            merge_text=merge_text,
        )

    def __bool__(self) -> bool:
        return bool(self.parts) and any(self.parts)


def _extract_parts(
    element: MultimodalContent | MultimodalContentConvertible,
    /,
) -> frozenlist[MultimodalContentElement]:
    match element:
        case MultimodalContent() as content:
            return content.parts

        case str() as text:
            return (TextContent(text=text),)

        case element:
            return (element,)


def _as_content(
    element: MultimodalContentConvertible,
    /,
) -> MultimodalContentElement:
    match element:
        case str() as text:
            return TextContent(text=text)

        case element:
            return element


def _is_media(
    element: MultimodalContentElement,
) -> bool:
    match element:
        case TextContent():
            return False

        case _:
            return True


def _is_artifact(  # noqa: PLR0911
    element: MultimodalContentElement,
) -> bool:
    match element:
        case TextContent():
            return False

        case ImageURLContent():
            return False

        case ImageBase64Content():
            return False

        case AudioURLContent():
            return False

        case AudioBase64Content():
            return False

        case VideoURLContent():
            return False

        case VideoBase64Content():
            return False

        case _:
            return True


def _as_string(  # noqa: PLR0911
    element: MultimodalContentElement,
) -> str:
    match element:
        case TextContent() as text:
            return text.text

        case ImageURLContent() as image_url:
            return f"![{image_url.image_description or 'IMAGE'}]({image_url.image_url})"

        case ImageBase64Content() as image_base64:
            # we might want to use base64 content directly, but it would make a lot of tokens...
            return f"![{image_base64.image_description or 'IMAGE'}]()"

        case AudioURLContent() as audio_url:
            return f"![{audio_url.audio_transcription or 'AUDIO'}]({audio_url.audio_url})"

        case AudioBase64Content() as audio_base64:
            # we might want to use base64 content directly, but it would make a lot of tokens...
            return f"![{audio_base64.audio_transcription or 'AUDIO'}]()"

        case VideoURLContent() as video_url:
            return f"![{video_url.video_transcription or 'VIDEO'}]({video_url.video_url})"

        case VideoBase64Content() as video_base64:
            # we might want to use base64 content directly, but it would make a lot of tokens...
            return f"![{video_base64.video_transcription or 'VIDEO'}]()"

        case DataModel() as model:
            return str(model)


def _merge_texts(
    *elements: MultimodalContentElement,
) -> Iterable[MultimodalContentElement]:
    if len(elements) <= 1:
        return elements

    result: list[MultimodalContentElement] = []
    iterator: Iterator[MultimodalContentElement] = iter(elements)
    last_text_element: TextContent | None = None
    while element := next(iterator, None):
        match element:
            case TextContent() as text if not text.meta:  # do not merge texts with metadata
                if last_text := last_text_element:
                    last_text_element = TextContent(text=last_text.text + text.text)

                else:
                    last_text_element = text

            case other:
                if last_text := last_text_element:
                    result.append(last_text)
                    last_text_element = None

                result.append(other)

    if last_text := last_text_element:
        result.append(last_text)

    return result



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.score import CommonScoreModel
from draive.generation import generate_model

__all__ = [
    "text_relevance_evaluator",
]


INSTRUCTION: str = """\
You will be given a reference text and a compared text based on the reference text.
Your task is to rate the compared text using only the Relevance metric, \
which is described in the Evaluation Criteria.
Please make sure you read and understand these instructions very carefully.
Keep this document open while reviewing, and refer to it as needed.

Evaluation Criteria:
Relevance (0.0-4.0) - selection of important content from the reference text.
The compared text should include only important information from the reference text.
Annotators should penalize compared texts that contain redundancies and excess information.

Rating Scale:
0.0: Very low relevance - the text contains mostly irrelevant or redundant information.
1.0: Low relevance - the text includes some important points but has \
significant irrelevant content.
2.0: Moderate relevance - the text covers most important points but includes \
some unnecessary information.
3.0: Good relevance - the text focuses on important information with minor inclusions \
of less relevant content.
4.0: Excellent relevance - the text precisely captures only the most important information \
from the reference text.

Evaluation Steps:
1. Read the compared text and the reference text carefully.
2. Compare the compared text to the reference text and identify \
the main points of the reference text.
3. Assess how well the compared text covers the main points of the reference text, \
and note any irrelevant or redundant information it contains.
4. Assign a relevance score from 0.0 to 4.0 based on the provided criteria.

Important: The score must be a decimal number from 0.0 to 4.0. 4.0 is the maximum, \
do not exceed this value.
"""


INPUT_TEMPLATE: str = """
<REFERENCE_TEXT>
{reference}
</REFERENCE_TEXT>

<COMPARED_TEXT>
{compared}
</COMPARED_TEXT>
"""


@evaluator(name="text_relevance")
async def text_relevance_evaluator(
    compared: str,
    /,
    reference: str,
) -> EvaluationScore:
    if not compared:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference text was empty!",
        )

    score: CommonScoreModel = await generate_model(
        CommonScoreModel,
        instruction=INSTRUCTION,
        input=INPUT_TEMPLATE.format(
            reference=reference,
            compared=compared,
        ),
        examples=[
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "The sun is the star at the center of our solar system. "
                        "It provides light and heat to Earth."
                    ),
                    compared=(
                        "Stars twinkle in the night sky. Some people believe in astrology. "
                        "The moon orbits the Earth. Astronauts have been to space. "
                        "Solar panels use energy from the sun."
                    ),
                ),
                CommonScoreModel(score=0.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Elephants are the largest land animals. They have long trunks and tusks. "
                        "Elephants live in herds and are known for their intelligence."
                    ),
                    compared=(
                        "Elephants are very big animals. They use their trunks to grab food "
                        "and water. Elephants live together in groups. They're smart and have "
                        "good memories. Some people ride elephants in zoos, "
                        "but this can be harmful to the animals."
                    ),
                ),
                CommonScoreModel(score=2.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Bicycles are a popular mode of transportation. They are eco-friendly "
                        "and provide exercise. However, cyclists need to follow "
                        "traffic rules for safety."
                    ),
                    compared=(
                        "Bicycles are widely used for travel. "
                        "They don't pollute and help people stay fit. "
                        "Cyclists must obey traffic laws to stay safe."
                    ),
                ),
                CommonScoreModel(score=4.0),
            ),
        ],
    )

    return score.normalized(divider=4)



from draive.parameters import DataModel

__all__ = [
    "VideoBase64Content",
    "VideoContent",
    "VideoURLContent",
]


class VideoURLContent(DataModel):
    mime_type: str | None = None
    video_url: str
    video_transcription: str | None = None
    meta: dict[str, str | float | int | bool | None] | None = None


class VideoBase64Content(DataModel):
    mime_type: str | None = None
    video_base64: str
    video_transcription: str | None = None
    meta: dict[str, str | float | int | bool | None] | None = None


VideoContent = VideoURLContent | VideoBase64Content



from draive.gemini.client import GeminiClient
from draive.gemini.config import GeminiConfig, GeminiEmbeddingConfig
from draive.gemini.embedding import gemini_embed_text
from draive.gemini.errors import GeminiException
from draive.gemini.lmm import gemini_lmm_invocation
from draive.gemini.tokenization import gemini_tokenize_text

__all__ = [
    "gemini_embed_text",
    "gemini_lmm_invocation",
    "gemini_tokenize_text",
    "GeminiClient",
    "GeminiConfig",
    "GeminiEmbeddingConfig",
    "GeminiException",
]



import sys
from abc import ABCMeta
from collections.abc import Callable
from copy import deepcopy
from dataclasses import fields as dataclass_fields
from dataclasses import is_dataclass
from types import GenericAlias
from typing import (
    Any,
    ClassVar,
    Generic,
    ParamSpec,
    Self,
    TypeVar,
    TypeVarTuple,
    cast,
    dataclass_transform,
    final,
    get_origin,
    overload,
)
from weakref import WeakValueDictionary

from draive.parameters.annotations import (
    ParameterDefaultFactory,
    allows_missing,
    object_annotations,
)
from draive.parameters.basic import BasicValue
from draive.parameters.path import ParameterPath
from draive.parameters.specification import (
    ParameterSpecification,
    parameter_specification,
)
from draive.parameters.validation import (
    ParameterValidationContext,
    ParameterValidationError,
    ParameterValidator,
    ParameterVerifier,
    parameter_validator,
)
from draive.utils import MISSING, Missing, freeze, is_missing, not_missing

__all__ = [
    "Field",
    "ParametrizedData",
]


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default: Value | Missing = MISSING,
    validator: ParameterValidator[Value] | Missing = MISSING,
    converter: Callable[[Value], BasicValue] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default_factory: Callable[[], Value] | Missing = MISSING,
    validator: ParameterValidator[Value] | Missing = MISSING,
    converter: Callable[[Value], BasicValue] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default: Value | Missing = MISSING,
    verifier: ParameterVerifier[Value] | None = None,
    converter: Callable[[Value], BasicValue] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default_factory: Callable[[], Value] | Missing = MISSING,
    verifier: ParameterVerifier[Value] | None = None,
    converter: Callable[[Value], BasicValue] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


def Field[Value](  # noqa: PLR0913 # Ruff - noqa: B008
    *,
    aliased: str | None = None,
    description: str | None = None,
    default: Value | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    validator: ParameterValidator[Value] | Missing = MISSING,
    verifier: ParameterVerifier[Value] | None = None,
    converter: Callable[[Value], BasicValue] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value:  # it is actually a DataField, but type checker has to be fooled
    assert (  # nosec: B101
        is_missing(default_factory) or is_missing(default)
    ), "Can't specify both default value and factory"
    assert (  # nosec: B101
        is_missing(validator) or verifier is None
    ), "Can't specify both validator and verifier"

    return cast(
        Value,
        DataField(
            aliased=aliased,
            description=description,
            default=default,
            default_factory=default_factory,
            validator=validator,
            verifier=verifier,
            converter=converter,
            specification=specification,
        ),
    )


@final
class DataField:
    def __init__(  # noqa: PLR0913
        self,
        aliased: str | None = None,
        description: str | None = None,
        default: Any | Missing = MISSING,
        default_factory: ParameterDefaultFactory[Any] | Missing = MISSING,
        validator: ParameterValidator[Any] | Missing = MISSING,
        verifier: ParameterVerifier[Any] | None = None,
        converter: Callable[[Any], BasicValue] | Missing = MISSING,
        specification: ParameterSpecification | Missing = MISSING,
    ) -> None:
        self.aliased: str | None = aliased
        self.description: str | None = description
        self.default: Any | Missing = default
        self.default_factory: Callable[[], Any] | Missing = default_factory
        self.validator: ParameterValidator[Any] | Missing = validator
        self.verifier: ParameterVerifier[Any] | None = verifier
        self.converter: Callable[[Any], BasicValue] | Missing = converter
        self.specification: ParameterSpecification | Missing = specification


@final
class DataParameter:
    @classmethod
    def of(  # noqa: PLR0913
        cls,
        annotation: Any,
        /,
        name: str,
        default: Any,
        prepare_specification: bool,
        type_arguments: dict[str, Any],
        globalns: dict[str, Any],
        localns: dict[str, Any] | None,
        recursion_guard: frozenset[type[Any]],
    ) -> Self:
        match default:
            case DataField() as data_field:
                return cls(
                    name=name,
                    aliased=data_field.aliased,
                    description=data_field.description,
                    annotation=annotation,
                    default=data_field.default,
                    default_factory=data_field.default_factory,
                    allows_missing=allows_missing(
                        annotation,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    ),
                    validator=data_field.validator
                    if not_missing(data_field.validator)
                    else parameter_validator(
                        annotation,
                        verifier=data_field.verifier,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    ),
                    converter=data_field.converter,
                    specification=(
                        data_field.specification
                        if not_missing(data_field.specification)
                        else parameter_specification(
                            annotation,
                            description=data_field.description,
                            type_arguments=type_arguments,
                            globalns=globalns,
                            localns=localns,
                            recursion_guard=recursion_guard,
                        )
                    )
                    if prepare_specification
                    else MISSING,
                )

            case default:
                return cls(
                    name=name,
                    aliased=None,
                    description=None,
                    annotation=annotation,
                    default=default,
                    default_factory=MISSING,
                    allows_missing=allows_missing(
                        annotation,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    ),
                    validator=parameter_validator(
                        annotation,
                        verifier=None,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    ),
                    converter=MISSING,
                    specification=parameter_specification(
                        annotation,
                        description=None,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    )
                    if prepare_specification
                    else MISSING,
                )

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        aliased: str | None,
        description: str | None,
        annotation: Any,
        default: Any | Missing,
        default_factory: ParameterDefaultFactory[Any] | Missing,
        allows_missing: bool,
        validator: ParameterValidator[Any],
        converter: Callable[[Any], BasicValue] | Missing,
        specification: ParameterSpecification | Missing,
    ) -> None:
        self.name: str = name
        self.aliased: str | None = aliased
        self.description: str | None = description
        self.annotation: Any = annotation
        self.default_value: Callable[[], Any | Missing]
        if not_missing(default_factory):
            self.default_value = default_factory

        elif not_missing(default):
            self.default_value = lambda: default

        else:
            self.default_value = lambda: MISSING

        self.has_default: bool = not_missing(default_factory) or not_missing(default)
        self.allows_missing: bool = allows_missing
        self.validator: ParameterValidator[Any] = validator
        self.converter: Callable[[Any], BasicValue] | Missing = converter
        self.specification: ParameterSpecification | Missing = specification

        freeze(self)

    def validated(
        self,
        value: Any,
        /,
        context: ParameterValidationContext,
    ) -> Any:
        if is_missing(value):
            if self.has_default:
                return self.validator(self.default_value(), context.appending_path(f".{self.name}"))

            elif self.allows_missing:
                return MISSING

            else:
                raise ParameterValidationError.missing(
                    context=context.appending_path(f".{self.name}")
                )

        else:
            return self.validator(value, context.appending_path(f".{self.name}"))


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
    field_specifiers=(Field,),
)
class ParametrizedDataMeta(ABCMeta):
    def __new__(  # noqa: PLR0913
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        type_base: type[Any] | None = None,
        type_parameters: tuple[TypeVar | ParamSpec | TypeVarTuple, ...] | None = None,
        type_arguments: tuple[type[Any], ...] | None = None,
        **kwargs: Any,
    ) -> Any:
        data_type = super().__new__(
            cls,
            name,
            bases,
            namespace,
            **kwargs,
        )

        data_type.__TYPE_BASE__ = type_base or data_type  # pyright: ignore[reportAttributeAccessIssue]
        data_type.__TYPE_ARGUMENTS__ = type_arguments or ()  # pyright: ignore[reportAttributeAccessIssue]

        if not bases:  # handle base class
            return data_type

        globalns: dict[str, Any]
        if data_type.__module__:
            globalns = sys.modules.get(data_type.__module__).__dict__

        else:
            globalns = {}

        localns: dict[str, Any] = {data_type.__name__: data_type}
        recursion_guard: frozenset[type[Any]] = frozenset({data_type})
        resolved_type_parameters: dict[str, Any] = {
            parameter.__name__: argument
            for (parameter, argument) in zip(
                type_parameters or (),
                type_arguments or (),
                strict=False,
            )
        }
        parameters: dict[str, DataParameter] = {}
        prepare_specification: bool = hasattr(
            data_type,
            "__PARAMETERS_SPECIFICATION__",
        )
        properties_specification: dict[str, ParameterSpecification] = {}
        aliased_required_specification: list[str] = []

        for key, annotation in object_annotations(
            data_type,
            globalns,
            localns,
        ).items():
            # do not include ClassVars and private or dunder items
            if ((get_origin(annotation) or annotation) is ClassVar) or key.startswith("__"):
                continue

            parameter: DataParameter = DataParameter.of(
                annotation,
                name=key,
                default=getattr(data_type, key, MISSING),
                # prepare specification only for the data models
                prepare_specification=prepare_specification,
                type_arguments=resolved_type_parameters,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )
            parameters[key] = parameter

            if prepare_specification:  # check and use specification when needed
                if not_missing(parameter.specification):
                    properties_specification[parameter.name] = parameter.specification

                    if not (parameter.has_default or parameter.allows_missing):
                        aliased_required_specification.append(key)

                else:
                    # if any parameter does not have specification then whole type does not have one
                    raise TypeError("Property %s of %s, has no specification!", key, annotation)

        data_type.__PARAMETERS__ = parameters  # pyright: ignore[reportAttributeAccessIssue]
        data_type.__PARAMETERS_LIST__ = list(parameters.values())  # pyright: ignore[reportAttributeAccessIssue]

        if prepare_specification:
            if properties_specification:
                data_type.__PARAMETERS_SPECIFICATION__ = {  # pyright: ignore[reportAttributeAccessIssue]
                    "type": "object",
                    "properties": properties_specification,
                    "required": aliased_required_specification,
                }

            else:
                data_type.__PARAMETERS_SPECIFICATION__ = {  # pyright: ignore[reportAttributeAccessIssue]
                    "type": "object",
                    "additionalProperties": True,
                }

        data_type.__slots__ = frozenset(parameters.keys())  # pyright: ignore[reportAttributeAccessIssue]
        data_type.__match_args__ = data_type.__slots__  # pyright: ignore[reportAttributeAccessIssue]
        data_type._ = ParameterPath(root=data_type, parameter=data_type)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
        return data_type

    def __instancecheck__(
        cls,
        instance: Any,
    ) -> bool:
        # Avoid calling ABC _abc_subclasscheck | https://github.com/python/cpython/issues/92810
        if not hasattr(instance, "__PARAMETERS__"):
            return False  # early exit if not subclass of ParametrizedData

        return super().__instancecheck__(instance)

    def __subclasscheck__(
        cls,
        subclass: Any,
    ) -> bool:
        # Avoid calling ABC _abc_subclasscheck | https://github.com/python/cpython/issues/92810
        if not hasattr(subclass, "__PARAMETERS__"):
            return False  # early exit if not subclass of ParametrizedData

        # check for parametrized subtypes within matching parameters
        if cls.__TYPE_BASE__ == subclass.__TYPE_BASE__:  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
            # check for parametrized variant by checking parameters
            return all(
                issubclass(subclass_param, cls_param)
                for (cls_param, subclass_param) in zip(  # pyright: ignore[reportUnknownVariableType]
                    cls.__TYPE_ARGUMENTS__,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType, reportAttributeAccessIssue]
                    subclass.__TYPE_ARGUMENTS__,
                    strict=False,
                )
            )

        return super().__subclasscheck__(subclass)


_types_cache: WeakValueDictionary[
    tuple[
        Any,
        tuple[Any, ...],
    ],
    Any,
] = WeakValueDictionary()


class ParametrizedData(metaclass=ParametrizedDataMeta):
    _: ClassVar[Self]
    __TYPE_BASE__: ClassVar[type[ParametrizedDataMeta]]
    __TYPE_ARGUMENTS__: ClassVar[tuple[type[Any], ...]]
    __PARAMETERS__: ClassVar[dict[str, DataParameter]]
    __PARAMETERS_LIST__: ClassVar[list[DataParameter]]

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        if self.__PARAMETERS_LIST__:
            for parameter in self.__PARAMETERS_LIST__:
                parameter_context: ParameterValidationContext = ParameterValidationContext(
                    path=(self.__class__.__qualname__,)
                )
                object.__setattr__(
                    self,  # pyright: ignore[reportUnknownArgumentType]
                    parameter.name,
                    parameter.validated(
                        kwargs.get(
                            parameter.name,
                            kwargs.get(
                                parameter.aliased,
                                MISSING,
                            )
                            if parameter.aliased
                            else MISSING,
                        ),
                        context=parameter_context,
                    ),
                )

        else:
            for key, value in kwargs.items():
                object.__setattr__(
                    self,  # pyright: ignore[reportUnknownArgumentType]
                    key,
                    value,
                )

    def __class_getitem__(
        cls,
        type_argument: tuple[type[Any], ...] | type[Any],
    ) -> type[Self]:
        assert Generic in cls.__bases__, "Can't specialize non generic type!"  # nosec: B101

        type_arguments: tuple[type[Any], ...]
        match type_argument:
            case [*arguments]:
                type_arguments = tuple(arguments)

            case argument:
                type_arguments = (argument,)

        if any(isinstance(argument, TypeVar) for argument in type_arguments):
            # if we got unfinished type treat it as an alias
            return cast(type[Self], GenericAlias(cls, type_arguments))

        assert len(type_arguments) == len(  # nosec: B101
            cls.__type_params__
        ), "Type arguments count has to match parameters count"

        # TODO: add UnionType ordering for equal types
        if cached := _types_cache.get((cls, type_arguments)):
            return cached

        parameter_names: str = ",".join(
            getattr(
                argument,
                "__name__",
                str(argument),
            )
            for argument in type_arguments
        )
        name: str = f"{cls.__name__}[{parameter_names}]"
        bases: tuple[type[Self]] = (cls,)

        parametrized_type: type[Self] = ParametrizedDataMeta.__new__(
            cls=cls.__class__,
            name=name,
            bases=bases,
            namespace={
                "__module__": cls.__module__,
            },
            type_base=cls,
            type_parameters=cls.__type_params__,  # pyright: ignore[reportArgumentType]
            type_arguments=type_arguments,
        )
        _types_cache[(cls, type_arguments)] = parametrized_type
        return parametrized_type

    @classmethod
    def path[Parameter](
        cls,
        path: Parameter,
        /,
    ) -> ParameterPath[Self, Parameter]:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        return cast(ParameterPath[Self, Parameter], path)

    @classmethod
    def validator(
        cls,
        /,
        value: Any,
        context: ParameterValidationContext,
    ) -> Self:
        match value:
            case valid if isinstance(valid, cls):
                return valid

            case {**values}:
                return cls(**values)

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=cls,
                    received=value,
                    context=context,
                )

    @classmethod
    def from_dict(
        cls,
        value: dict[str, Any],
        /,
    ) -> Self:
        return cls(**value)

    def as_dict(
        self,
        aliased: bool = True,
    ) -> dict[str, Any]:
        return _data_dict(
            self,
            aliased=aliased,
            converter=None,
        )

    # TODO: find a way to generate signature similar to dataclass __init__
    # or try using ParameterPath if found a way to generate setters
    def updated(
        self,
        /,
        **parameters: Any,
    ) -> Self:
        if not parameters:
            return self

        updated: Self = self.__new__(self.__class__)
        for parameter in self.__PARAMETERS_LIST__:
            parameter_context: ParameterValidationContext = ParameterValidationContext(
                path=(self.__class__.__qualname__,),
            )
            validated_value: Any
            if parameter.name in parameters:
                validated_value = parameter.validated(
                    parameters[parameter.name],
                    context=parameter_context,
                )

            elif parameter.aliased and parameter.aliased in parameters:
                validated_value = parameter.validated(
                    parameters[parameter.aliased],
                    context=parameter_context,
                )

            else:  # no need to validate again when reusing current value
                validated_value = object.__getattribute__(
                    self,
                    parameter.name,
                )

            object.__setattr__(
                updated,
                parameter.name,
                validated_value,
            )

        return updated

    def updating[Value](
        self,
        path: ParameterPath[Self, Value] | Value,
        /,
        value: Value,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        return cast(ParameterPath[Self, Value], path)(self, updated=value)

    def updating_parameter(
        self,
        name: str,
        /,
        value: Any,
    ) -> Self:
        assert (  # nosec: B101
            name in self.__PARAMETERS__
        ), f"Parameter {name} does not exist in {self.__class__.__qualname__}"

        updated: Self = self.__new__(self.__class__)
        for parameter in self.__PARAMETERS_LIST__:
            parameter_context: ParameterValidationContext = ParameterValidationContext(
                path=(self.__class__.__qualname__,),
            )
            validated_value: Any
            if parameter.name == name:
                validated_value = parameter.validated(
                    value,
                    context=parameter_context,
                )

            elif parameter.aliased and parameter.aliased == name:
                validated_value = parameter.validated(
                    value,
                    context=parameter_context,
                )

            else:  # no need to validate again when reusing current value
                validated_value = object.__getattribute__(
                    self,
                    parameter.name,
                )

            object.__setattr__(
                updated,
                parameter.name,
                validated_value,
            )

        return updated

    def __str__(self) -> str:
        return _data_str(self, aliased=True, converter=None).strip()

    def __repr__(self) -> str:
        return str(self.as_dict())

    def __eq__(self, other: Any) -> bool:
        if other.__class__ != self.__class__:
            return False

        return all(
            getattr(self, key, MISSING) == getattr(other, key, MISSING)
            for key in self.__PARAMETERS__.keys()
        )

    def __getitem__(
        self,
        name: str,
    ) -> Any | Missing:
        return vars(self).get(name, MISSING)

    def __setattr__(
        self,
        __name: str,
        __value: Any,
    ) -> None:
        raise AttributeError(f"{self.__class__.__qualname__} is frozen and can't be modified")

    def __delattr__(
        self,
        __name: str,
    ) -> None:
        raise AttributeError(f"{self.__class__.__qualname__} is frozen and can't be modified")

    def __copy__(self) -> Self:
        return self.__class__(**vars(self))

    def __deepcopy__(
        self,
        memo: dict[int, Any] | None,
    ) -> Self:
        copy: Self = self.__class__(
            **{
                key: deepcopy(
                    value,
                    memo,
                )
                for key, value in vars(self).items()
            }
        )
        return copy


# based on python dataclass asdict but simplified
def _data_dict(  # noqa: PLR0911
    data: Any,
    /,
    aliased: bool,
    converter: Callable[[Any], BasicValue] | None,
) -> Any:
    # use converter if able
    if converter := converter:
        return converter(data)

    match data:
        case str() | None | int() | float() | bool():
            return data  # use basic value types as they are

        case parametrized_data if hasattr(parametrized_data, "__PARAMETERS_LIST__"):
            # convert parametrized data to dict
            if not parametrized_data.__PARAMETERS_LIST__:  # if base - use all variables
                return cast(dict[str, Any], vars(parametrized_data))

            elif aliased:  # alias if needed
                return {
                    field.aliased or field.name: _data_dict(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    )
                    for field in parametrized_data.__PARAMETERS_LIST__
                }

            else:
                return {
                    field.name: _data_dict(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    )
                    for field in parametrized_data.__PARAMETERS_LIST__
                }

        case {**elements}:  # replace mapping with dict
            return {
                key: _data_dict(
                    value,
                    aliased=aliased,
                    converter=None,
                )
                for key, value in elements.items()
            }

        case [*values]:  # replace sequence with list
            return [
                _data_dict(
                    value,
                    aliased=aliased,
                    converter=None,
                )
                for value in values
            ]

        case dataclass if is_dataclass(dataclass):
            return {
                f.name: _data_dict(
                    getattr(dataclass, f.name),
                    aliased=aliased,
                    converter=None,
                )
                for f in dataclass_fields(dataclass)
            }

        case other:  # for other types use deepcopy
            return deepcopy(other)


def _data_str(  # noqa: PLR0911, PLR0912, C901
    data: Any,
    /,
    aliased: bool,
    converter: Callable[[Any], BasicValue] | None,
) -> str:
    # use converter if able
    if converter := converter:
        return _data_str(
            converter(data),
            aliased=aliased,
            converter=None,
        )

    match data:
        case str() as string:
            return string

        case None | int() | float() | bool():
            return str(data)

        case parametrized_data if hasattr(parametrized_data, "__PARAMETERS_LIST__"):
            # convert parametrized data to dict
            if not parametrized_data.__PARAMETERS_LIST__:  # if base - use all variables
                string: str = ""
                for key, value in vars(parametrized_data).items():  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                    element: str = _data_str(
                        value,
                        aliased=aliased,
                        converter=None,
                    ).replace("\n", "\n  ")

                    string += f"\n{key}: {element}"

                return string

            elif aliased:  # alias if needed
                string: str = ""
                for field in parametrized_data.__PARAMETERS_LIST__:
                    element: str = _data_str(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    ).replace("\n", "\n  ")

                    string += f"\n{field.aliased or field.name}: {element}"

                return string

            else:
                string: str = ""
                for field in parametrized_data.__PARAMETERS_LIST__:
                    element: str = _data_str(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    ).replace("\n", "\n  ")

                    string += f"\n{field.name}: {element}"

                return string

        case {**elements}:  # replace mapping with dict
            string: str = ""
            for key, value in elements.items():
                element: str = _data_str(
                    value,
                    aliased=aliased,
                    converter=None,
                ).replace("\n", "\n  ")

                string += f"\n{key}: {element}"

            return string

        case [*values]:  # replace sequence with list
            string: str = ""
            for value in values:
                element: str = (
                    _data_str(
                        value,
                        aliased=aliased,
                        converter=None,
                    )
                    .replace("\n", "\n  ")
                    .strip()
                )

                string += f"\n- {element}"

            return string

        case dataclass if is_dataclass(dataclass):
            string: str = ""
            for field in dataclass_fields(dataclass):
                element: str = _data_str(
                    getattr(parametrized_data, field.name),
                    aliased=aliased,
                    converter=None,
                ).replace("\n", "\n  ")

                string += f"\n{field.name}: {element}"

            return string

        case other:  # for other types use its str
            return str(other)



from collections.abc import Callable
from typing import Any, Literal

from draive import MISSING, Field, Missing, State


def invalid(value: str) -> None:
    if value != "valid":
        raise ValueError()


class ExampleNestedState(State):
    string: str = "default"
    more: list[int] | None = None


def dummy_callable() -> None:
    pass


class ExampleState(State):
    string: str = "default"
    number: int = Field(aliased="alias", description="description", default=1)
    none_default: int | None = Field(default=None)
    value_default: int = Field(default=9)
    invalid: str = Field(verifier=invalid, default="valid")
    nested: ExampleNestedState = Field(aliased="answer", default=ExampleNestedState())
    full: Literal["A", "B"] | list[int] | str | bool | None = Field(
        aliased="all",
        description="complex",
        default="",
    )
    function: Callable[..., Any] = dummy_callable


# TODO: prepare extensive tests
def test_validated_passes_with_valid_values() -> None:
    values_dict: dict[str, Any] = {
        "string": "value",
        "alias": 42,
        "none_default": 1,
        "value_default": 8,
        "invalid": "valid",
        "answer": {"string": "value", "more": None},
        "all": True,
        "function": dummy_callable,
    }
    model: ExampleState = ExampleState(**values_dict)
    assert model.as_dict() == values_dict


def test_validated_passes_with_default_values() -> None:
    ExampleState()


class MissingState(State):
    value: Any | Missing


missingStateInstance: MissingState = MissingState(
    value=MISSING,
)
missingStateDict: dict[str, Any] = {"value": MISSING}


def test_missing_encoding() -> None:
    assert missingStateInstance.as_dict() == missingStateDict


def test_missing_decoding() -> None:
    assert MissingState.from_dict(missingStateDict) == missingStateInstance


class BasicsState(State):
    string: str
    string_list: list[str]
    integer: int
    integer_or_float_list: list[int | float]
    floating: float
    floating_dict: dict[str, float]
    optional: str | None
    none: None


basicStateInstance: BasicsState = BasicsState(
    string="test",
    string_list=["basic", "list"],
    integer=42,
    integer_or_float_list=[12, 3.14, 7],
    floating=9.99,
    floating_dict={"a": 65, "b": 66.0, "c": 67.5},
    optional="some",
    none=None,
)
basicStateDict: dict[str, Any] = {
    "string": "test",
    "string_list": ["basic", "list"],
    "integer": 42,
    "integer_or_float_list": [12, 3.14, 7],
    "floating": 9.99,
    "floating_dict": {"a": 65, "b": 66.0, "c": 67.5},
    "optional": "some",
    "none": None,
}


def test_basic_encoding() -> None:
    assert basicStateInstance.as_dict() == basicStateDict


def test_basic_decoding() -> None:
    assert BasicsState.from_dict(basicStateDict) == basicStateInstance


def test_type_parametrized_state() -> None:
    class Parameter(State):
        value: int

    class Parametrized[Value](State):
        value: Value

    ParametrizedAlias = Parametrized[str]
    ParametrizedParameterAlias = Parametrized[Parameter]

    # those should only instantiate without any issues
    ParametrizedAlias(value="")
    ParametrizedParameterAlias(value=Parameter(value=42))



from collections.abc import Iterable, Sequence
from typing import Any

from draive.generation.text.state import TextGeneration
from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox
from draive.scope import ctx
from draive.types import MultimodalContent, MultimodalContentConvertible

__all__ = [
    "generate_text",
]


async def generate_text(
    *,
    instruction: Instruction | str,
    input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
    tools: Toolbox | Sequence[AnyTool] | None = None,
    examples: Iterable[tuple[MultimodalContent | MultimodalContentConvertible, str]] | None = None,
    **extra: Any,
) -> str:
    return await ctx.state(TextGeneration).generate(
        instruction=instruction,
        input=input,
        tools=tools,
        examples=examples,
        **extra,
    )



from draive import xml_tag, xml_tags


def test_returns_none_with_empty():
    source: str = ""

    assert xml_tag("test", source=source) is None


def test_returns_none_without_tag():
    source: str = "Lorem ipsum"

    assert xml_tag("test", source=source) is None


def test_returns_none_with_other_tag():
    source: str = "<other>Lorem ipsum</other>"

    assert xml_tag("test", source=source) is None


def test_returns_none_with_closing_tag():
    source: str = "Lorem ipsum</test>"

    assert xml_tag("test", source=source) is None


def test_returns_none_with_reversed_tags():
    source: str = "</test>Lorem ipsum<test>"

    assert xml_tag("test", source=source) is None


def test_returns_none_without_closing_tag():
    source: str = "<test>Lorem ipsum"

    assert xml_tag("test", source=source) is None


def test_returns_none_with_malformed_opening_tag():
    source: str = "<testx>Lorem ipsum</test>"

    assert xml_tag("test", source=source) is None


def test_returns_none_with_malformed_closing_tag():
    source: str = "<test>Lorem ipsum</testx>"

    assert xml_tag("test", source=source) is None


def test_returns_content_with_valid_tag():
    source: str = "<test>Lorem ipsum</test>"

    assert xml_tag("test", source=source) == "Lorem ipsum"


def test_returns_content_with_surrounded_tag():
    source: str = "Lorem<test>Lorem ipsum</test>ipsum"

    assert xml_tag("test", source=source) == "Lorem ipsum"


def test_returns_content_with_opening_tag_containing_extras():
    source: str = "<test extra items=here>Lorem ipsum</test>"

    assert xml_tag("test", source=source) == "Lorem ipsum"


def test_returns_first_content_with_multiple_tags():
    source: str = "<test>Lorem ipsum</test><test>Other</test>"

    assert xml_tag("test", source=source) == "Lorem ipsum"


def test_returns_nested_content_with_multiple_nested_tags():
    source: str = "<test>Other<test>Lorem ipsum</test></test>"

    assert xml_tag("test", source=source) == "Lorem ipsum"


def test_returns_nested_content_with_fake_tags():
    source: str = "<test>Lorem<ipsum</test>"

    assert xml_tag("test", source=source) == "Lorem<ipsum"


def test_returns_nested_content_with_other_tags():
    source: str = "<other>Other<more><test>Lorem</more>ipsum</test></other>"

    assert xml_tag("test", source=source) == "Lorem</more>ipsum"


def test_returns_content_with_multiple_tags():
    source: str = "<test>Lorem ipsum</test><test>Dolor</test><test>Sit amet</test>"
    contents: list[str] = list(xml_tags("test", source=source))

    assert contents == ["Lorem ipsum", "Dolor", "Sit amet"]


def test_skips_content_from_different_tags():
    source: str = "<test>Lorem ipsum</test><other>Other</other><more><test>Dolor</test>"
    contents: list[str] = list(xml_tags("test", source=source))

    assert contents == ["Lorem ipsum", "Dolor"]



from draive.agents import (
    Agent,
    AgentError,
    AgentException,
    AgentInvocation,
    AgentMessage,
    AgentNode,
    AgentOutput,
    AgentWorkflow,
    AgentWorkflowInput,
    AgentWorkflowInvocation,
    AgentWorkflowOutput,
    agent,
    workflow,
)
from draive.conversation import (
    Conversation,
    ConversationCompletion,
    ConversationMessage,
    ConversationMessageChunk,
    ConversationResponseStream,
    conversation_completion,
    lmm_conversation_completion,
)
from draive.embedding import (
    Embedded,
    ImageEmbedding,
    TextEmbedding,
    ValueEmbedder,
    embed_image,
    embed_images,
    embed_text,
    embed_texts,
)
from draive.generation import (
    ImageGeneration,
    ImageGenerator,
    ModelGeneration,
    ModelGenerator,
    ModelGeneratorDecoder,
    TextGeneration,
    TextGenerator,
    generate_image,
    generate_model,
    generate_text,
)
from draive.helpers import (
    ConstantMemory,
    VolatileAccumulativeMemory,
    VolatileMemory,
    VolatileVectorIndex,
    auto_retry,
    traced,
)
from draive.instructions import (
    Instruction,
    InstructionFetching,
    InstructionsRepository,
    MissingInstruction,
    instruction,
    instructions_file,
)
from draive.lmm import (
    LMM,
    Tool,
    ToolAvailabilityCheck,
    Toolbox,
    ToolContext,
    ToolException,
    ToolStatus,
    lmm_invocation,
    tool,
)
from draive.metrics import (
    Metric,
    MetricsTrace,
    MetricsTraceReport,
    MetricsTraceReporter,
    TokenUsage,
    metrics_log_reporter,
)
from draive.parameters import (
    Argument,
    BasicValue,
    DataModel,
    Field,
    ParameterDefaultFactory,
    ParameterPath,
    ParameterRequirement,
    ParameterValidationContext,
    ParameterValidationError,
    ParameterValidator,
    ParameterVerifier,
    State,
    Stateless,
)
from draive.scope import (
    ScopeDependencies,
    ScopeDependency,
    ScopeState,
    ctx,
)
from draive.similarity import (
    mmr_vector_similarity_search,
    vector_similarity_score,
    vector_similarity_search,
)
from draive.splitters import split_text
from draive.tokenization import TextTokenizer, Tokenization, count_text_tokens, tokenize_text
from draive.types import (
    JSON,
    AudioBase64Content,
    AudioContent,
    AudioURLContent,
    BasicMemory,
    ImageBase64Content,
    ImageContent,
    ImageURLContent,
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolResponse,
    Memory,
    MultimodalContent,
    MultimodalContentConvertible,
    MultimodalContentElement,
    RateLimitError,
    TextContent,
    VideoBase64Content,
    VideoContent,
    VideoURLContent,
    frozenlist,
)
from draive.utils import (
    MISSING,
    AsyncQueue,
    AsyncStream,
    Missing,
    always,
    async_always,
    async_noop,
    asynchronous,
    cache,
    freeze,
    getenv_bool,
    getenv_float,
    getenv_int,
    getenv_str,
    is_missing,
    load_env,
    markdown_block,
    markdown_blocks,
    noop,
    not_missing,
    setup_logging,
    split_sequence,
    with_timeout,
    xml_tag,
    xml_tags,
)

__all__ = [
    "agent",
    "Agent",
    "AgentError",
    "AgentException",
    "AgentInvocation",
    "AgentMessage",
    "AgentNode",
    "AgentOutput",
    "AgentWorkflow",
    "AgentWorkflowInput",
    "AgentWorkflowInvocation",
    "AgentWorkflowOutput",
    "always",
    "Argument",
    "async_always",
    "async_noop",
    "asynchronous",
    "AsyncQueue",
    "AsyncStream",
    "AudioBase64Content",
    "AudioContent",
    "AudioURLContent",
    "auto_retry",
    "BasicMemory",
    "BasicValue",
    "cache",
    "ConstantMemory",
    "conversation_completion",
    "Conversation",
    "ConversationCompletion",
    "ConversationMessage",
    "ConversationMessageChunk",
    "ConversationResponseStream",
    "count_text_tokens",
    "ctx",
    "DataModel",
    "embed_image",
    "embed_images",
    "embed_text",
    "embed_texts",
    "Embedded",
    "Field",
    "freeze",
    "frozenlist",
    "generate_image",
    "generate_model",
    "generate_text",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "ImageBase64Content",
    "ImageContent",
    "ImageEmbedding",
    "ImageGeneration",
    "ImageGenerator",
    "ImageURLContent",
    "instruction",
    "Instruction",
    "Instruction",
    "InstructionFetching",
    "instructions_file",
    "InstructionsRepository",
    "is_missing",
    "JSON",
    "lmm_conversation_completion",
    "lmm_invocation",
    "LMM",
    "LMMCompletion",
    "LMMCompletionChunk",
    "LMMContextElement",
    "LMMInput",
    "LMMOutputStream",
    "LMMOutputStreamChunk",
    "LMMToolRequest",
    "LMMToolResponse",
    "load_env",
    "markdown_block",
    "markdown_blocks",
    "Memory",
    "Metric",
    "metrics_log_reporter",
    "MetricsTrace",
    "MetricsTraceReport",
    "MetricsTraceReporter",
    "Missing",
    "MISSING",
    "MissingInstruction",
    "mmr_vector_similarity_search",
    "ModelGeneration",
    "ModelGenerator",
    "ModelGeneratorDecoder",
    "MultimodalContent",
    "MultimodalContentConvertible",
    "MultimodalContentElement",
    "noop",
    "not_missing",
    "ParameterDefaultFactory",
    "ParameterPath",
    "ParameterRequirement",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterValidator",
    "ParameterVerifier",
    "RateLimitError",
    "ScopeDependencies",
    "ScopeDependency",
    "ScopeState",
    "setup_logging",
    "split_sequence",
    "split_text",
    "State",
    "Stateless",
    "TextContent",
    "TextEmbedding",
    "TextGeneration",
    "TextGenerator",
    "TextTokenizer",
    "Tokenization",
    "tokenize_text",
    "TokenUsage",
    "tool",
    "Tool",
    "ToolAvailabilityCheck",
    "Toolbox",
    "ToolContext",
    "ToolException",
    "ToolException",
    "ToolStatus",
    "traced",
    "ValueEmbedder",
    "vector_similarity_score",
    "vector_similarity_search",
    "VideoBase64Content",
    "VideoContent",
    "VideoURLContent",
    "VolatileAccumulativeMemory",
    "VolatileMemory",
    "VolatileVectorIndex",
    "with_timeout",
    "workflow",
    "xml_tag",
    "xml_tags",
]



from typing import Any

from openai.types.image import Image

from draive.openai.client import OpenAIClient
from draive.openai.config import OpenAIImageGenerationConfig
from draive.openai.errors import OpenAIException
from draive.scope import ctx
from draive.types import ImageBase64Content, ImageContent, ImageURLContent

__all__ = [
    "openai_generate_image",
]


async def openai_generate_image(
    *,
    instruction: str,
    **extra: Any,
) -> ImageContent:
    client: OpenAIClient = ctx.dependency(OpenAIClient)
    config: OpenAIImageGenerationConfig = ctx.state(OpenAIImageGenerationConfig).updated(**extra)
    with ctx.nested("openai_generate_image", metrics=[config]):
        image: Image = await client.generate_image(
            config=config,
            instruction=instruction,
        )
        if url := image.url:
            return ImageURLContent(image_url=url)

        elif b64data := image.b64_json:
            return ImageBase64Content(image_base64=b64data)

        else:
            raise OpenAIException("Invalid OpenAI response - missing image content")



from collections.abc import Sequence
from typing import Any, Protocol, runtime_checkable

from draive.embedding.embedded import Embedded

__all__ = [
    "ValueEmbedder",
]


@runtime_checkable
class ValueEmbedder[Value](Protocol):
    async def __call__(
        self,
        values: Sequence[Value],
        **extra: Any,
    ) -> list[Embedded[Value]]: ...



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.score import CommonScoreModel
from draive.generation import generate_model

__all__ = [
    "text_conciseness_evaluator",
]


INSTRUCTION: str = """\
You will be given a reference text and a compared text based on the reference text.
Your task is to rate the compared text using only the Conciseness metric, \
which is described in the Evaluation Criteria.
Please make sure you read and understand these instructions very carefully.
Keep this document open while reviewing, and refer to it as needed.

Evaluation Criteria:
Conciseness (0.0-4.0) - the extent to which the compared text is brief and to the point \
while still covering all key information.
A concise compared text avoids unnecessary details and repetition.
Annotators should penalize compared texts that are overly verbose or include irrelevant information.

Rating Scale:
0.0: Very low conciseness - the text is excessively verbose with much irrelevant information.
1.0: Low conciseness - the text contains unnecessary details and some irrelevant information.
2.0: Moderate conciseness - the text is somewhat concise but could be more focused.
3.0: Good conciseness - the text is mostly concise with minimal unnecessary information.
4.0: Excellent conciseness - the text is highly concise, containing only essential information.

Evaluation Steps:
1. Read the derived text and the reference text carefully.
2. Compare the compared text to the reference text and identify the main \
points of the reference text.
3. Assess how well the compared text covers the main points of the reference text, \
and how much irrelevant or redundant information it contains.
4. Assign a conciseness score from 0.0 to 4.0 based on the provided criteria.

Important: The score must be a decimal number from 0.0 to 4.0. 4.0 is the maximum, \
do not exceed this value.
"""


INPUT_TEMPLATE: str = """
<REFERENCE_TEXT>
{reference}
</REFERENCE_TEXT>

<COMPARED_TEXT>
{compared}
</COMPARED_TEXT>
"""


@evaluator(name="text_conciseness")
async def text_conciseness_evaluator(
    compared: str,
    /,
    reference: str,
) -> EvaluationScore:
    if not compared:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference text was empty!",
        )

    score: CommonScoreModel = await generate_model(
        CommonScoreModel,
        instruction=INSTRUCTION,
        input=INPUT_TEMPLATE.format(
            reference=reference,
            compared=compared,
        ),
        examples=[
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Solar energy is a renewable energy source that is gaining popularity. "
                        "Solar panels convert sunlight into electricity. "
                        "This technology is environmentally friendly and can reduce electricity "
                        "bills. However,installing solar panels requires an initial investment and "
                        "is dependent on weather conditions."
                    ),
                    compared=(
                        "Did you know that solar energy is becoming super popular these days? "
                        "It's this amazing, eco-friendly way to make electricity using "
                        "the sun's rays. People are getting really excited about it! Basically, "
                        "you put these special panels on your roof, and they soak up the sunlight "
                        "like a sponge. Then, through some pretty cool science stuff, "
                        "they turn that sunlight into electricity you can use in your house. "
                        "It's pretty neat, right? And get this - it can actually help you save "
                        "money on your electricity bills in the long run. But here's the thing: "
                        "you've got to shell out some cash upfront to get those panels installed. "
                        "It's kind of like buying a fancy coffee machine - costs a bit at first, "
                        "but then you save on all those coffee shop visits."
                    ),
                ),
                CommonScoreModel(score=0.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Coffee is a popular beverage worldwide. "
                        "It's made from roasted coffee beans. Caffeine in coffee "
                        "can boost energy and alertness. However, excessive consumption may "
                        "lead to sleep issues."
                    ),
                    compared=(
                        "Coffee is a widely consumed beverage made from roasted coffee beans. "
                        "It contains caffeine, which can enhance energy and alertness. However, "
                        "drinking too much coffee may cause sleep problems. "
                        "People enjoy coffee for its taste and stimulating effects, but it's "
                        "important to consume it in moderation."
                    ),
                ),
                CommonScoreModel(score=2.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "The water cycle, also known as the hydrologic cycle, "
                        "describes the continuous movement of water within the Earth and "
                        "atmosphere. It involves processes such as evaporation, condensation, "
                        "precipitation, and runoff."
                    ),
                    compared=(
                        "The water cycle is the continuous movement of water on Earth. "
                        "It includes evaporation, condensation, precipitation, and runoff."
                    ),
                ),
                CommonScoreModel(score=4.0),
            ),
        ],
    )
    return score.normalized(divider=4)



import json
from collections.abc import AsyncGenerator, Sequence
from typing import Any, Literal, cast, overload

from draive.instructions import Instruction
from draive.lmm import LMMToolSelection, ToolSpecification
from draive.metrics import ArgumentsTrace, ResultTrace, TokenUsage
from draive.mistral.client import MistralClient
from draive.mistral.config import MistralChatConfig
from draive.mistral.errors import MistralException
from draive.mistral.models import ChatCompletionResponse, ChatMessage, ChatMessageResponse
from draive.scope import ctx
from draive.types import (
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
)

__all__ = [
    "mistral_lmm_invocation",
]


@overload
async def mistral_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def mistral_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def mistral_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def mistral_lmm_invocation(  # noqa: PLR0913
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    with ctx.nested(
        "mistral_lmm_invocation",
        metrics=[
            ArgumentsTrace.of(
                instruction=instruction,
                context=context,
                tools=tools,
                tool_selection=tool_selection,
                output=output,
                stream=stream,
                **extra,
            ),
        ],
    ):
        ctx.log_debug("Requested Mistral lmm")
        client: MistralClient = ctx.dependency(MistralClient)
        config: MistralChatConfig = ctx.state(MistralChatConfig).updated(**extra)
        ctx.record(config)

        match output:
            case "text":
                config = config.updated(response_format={"type": "text"})

            case "json":
                if tools:
                    ctx.log_warning(
                        "Attempting to use Mistral in JSON mode with tools which is not supported."
                        " Using text mode instead..."
                    )
                    config = config.updated(response_format={"type": "text"})

                else:
                    config = config.updated(response_format={"type": "json_object"})

        messages: list[ChatMessage] = [
            ChatMessage(
                role="system",
                content=Instruction.of(instruction).format(),
            ),
            *[_convert_context_element(element=element) for element in context],
        ]

        if stream:
            return ctx.stream(
                _chat_completion_stream(
                    client=client,
                    config=config,
                    messages=messages,
                    tools=tools,
                    tool_selection=tool_selection,
                ),
            )

        else:
            return await _chat_completion(
                client=client,
                config=config,
                messages=messages,
                tools=tools,
                tool_selection=tool_selection,
            )


def _convert_context_element(
    element: LMMContextElement,
) -> ChatMessage:
    match element:
        case LMMInput() as input:
            return ChatMessage(
                role="user",
                content=input.content.as_string(),
            )

        case LMMCompletion() as completion:
            return ChatMessage(
                role="assistant",
                content=completion.content.as_string(),
            )

        case LMMToolRequests() as tool_requests:
            return ChatMessage(
                role="assistant",
                content="",
                tool_calls=[
                    {
                        "id": request.identifier,
                        "type": "function",
                        "function": {
                            "name": request.tool,
                            "arguments": json.dumps(request.arguments),
                        },
                    }
                    for request in tool_requests.requests
                ],
            )

        case LMMToolResponse() as tool_response:
            return ChatMessage(
                role="tool",
                name=tool_response.tool,
                content=tool_response.content.as_string(),
            )


async def _chat_completion(
    *,
    client: MistralClient,
    config: MistralChatConfig,
    messages: list[ChatMessage],
    tools: Sequence[ToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> LMMOutput:
    completion: ChatCompletionResponse
    match tool_selection:
        case "auto":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                tools=cast(
                    list[dict[str, object]],
                    tools,
                ),
                tool_choice="auto",
            )
        case "none":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                tools=[],
                tool_choice="none",
            )

        case "required":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                tools=cast(
                    list[dict[str, object]],
                    tools,
                ),
                tool_choice="any",
            )

        case tool:
            assert tool in (tools or []), "Can't suggest a tool without using it"  # nosec: B101
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                tools=cast(
                    list[dict[str, object]],
                    [tool],  # mistral can't be suggested with concrete tool
                ),
                tool_choice="any",
            )

    if usage := completion.usage:
        ctx.record(
            TokenUsage.for_model(
                config.model,
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.completion_tokens,
            ),
        )

    if not completion.choices:
        raise MistralException("Invalid Mistral completion - missing messages!", completion)

    completion_message: ChatMessageResponse = completion.choices[0].message

    if (tool_calls := completion_message.tool_calls) and (tools := tools):
        ctx.record(ResultTrace.of(tool_calls))

        return LMMToolRequests(
            requests=[
                LMMToolRequest(
                    identifier=call.id,
                    tool=call.function.name,
                    arguments=json.loads(call.function.arguments)
                    if isinstance(call.function.arguments, str)
                    else call.function.arguments,
                )
                for call in tool_calls
            ]
        )

    elif message := completion_message.content:
        ctx.record(ResultTrace.of(message))
        match message:
            case str(content):
                return LMMCompletion.of(content)

            # API docs say that it can be only a string in response
            # however library model allows list as well
            case other:
                return LMMCompletion.of(*other)

    else:
        raise MistralException("Invalid Mistral completion", completion)


async def _chat_completion_stream(
    *,
    client: MistralClient,
    config: MistralChatConfig,
    messages: list[ChatMessage],
    tools: Sequence[ToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> AsyncGenerator[LMMOutputStreamChunk, None]:
    ctx.log_debug("Mistral streaming api is not supported yet, using regular response...")
    output: LMMOutput = await _chat_completion(
        client=client,
        config=config,
        messages=messages,
        tools=tools,
        tool_selection=tool_selection,
    )

    match output:
        case LMMCompletion() as completion:
            yield LMMCompletionChunk.of(completion.content)

        case other:
            yield other



from asyncio import gather
from collections.abc import Sequence
from itertools import chain
from random import uniform
from typing import Literal, Self, cast, final, overload

from openai import AsyncAzureOpenAI, AsyncOpenAI, AsyncStream
from openai import RateLimitError as OpenAIRateLimitError
from openai._types import NOT_GIVEN, NotGiven
from openai.types import Moderation, ModerationCreateResponse
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.create_embedding_response import CreateEmbeddingResponse
from openai.types.image import Image
from openai.types.images_response import ImagesResponse

from draive.openai.config import (
    OpenAIChatConfig,
    OpenAIEmbeddingConfig,
    OpenAIImageGenerationConfig,
)
from draive.scope import ScopeDependency
from draive.types import RateLimitError
from draive.utils import freeze, getenv_str, not_missing

__all__ = [
    "OpenAIClient",
]


@final
class OpenAIClient(ScopeDependency):
    @classmethod
    def prepare(cls) -> Self:
        return cls(
            base_url=getenv_str("OPENAI_BASE_URL"),
            api_key=getenv_str("AZURE_OPENAI_API_KEY") or getenv_str("OPENAI_API_KEY"),
            organization=getenv_str("OPENAI_ORGANIZATION"),
            azure_api_endpoint=getenv_str("AZURE_OPENAI_API_BASE"),
            azure_api_version=getenv_str("AZURE_OPENAI_API_VERSION"),
            azure_deployment=getenv_str("AZURE_OPENAI_DEPLOYMENT_NAME"),
        )

    def __init__(  # noqa: PLR0913
        self,
        base_url: str | None,
        api_key: str | None,
        organization: str | None = None,
        azure_api_endpoint: str | None = None,
        azure_api_version: str | None = None,
        azure_deployment: str | None = None,
    ) -> None:
        # if all AZURE settings were provided use it as provider
        if azure_api_endpoint and azure_deployment and azure_api_version:
            self._client: AsyncOpenAI = AsyncAzureOpenAI(
                api_key=api_key,
                azure_endpoint=azure_api_endpoint,
                azure_deployment=azure_deployment,
                api_version=azure_api_version,
                organization=organization,
            )
        # otherwise try using OpenAI
        else:
            self._client: AsyncOpenAI = AsyncOpenAI(
                base_url=base_url,
                api_key=api_key,
                organization=organization,
            )

        freeze(self)

    @overload
    async def chat_completion(
        self,
        *,
        config: OpenAIChatConfig,
        messages: list[ChatCompletionMessageParam],
        tools: list[ChatCompletionToolParam] | None = None,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        stream: Literal[True],
    ) -> AsyncStream[ChatCompletionChunk]: ...

    @overload
    async def chat_completion(
        self,
        *,
        config: OpenAIChatConfig,
        messages: list[ChatCompletionMessageParam],
        tools: list[ChatCompletionToolParam] | None = None,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
    ) -> ChatCompletion: ...

    async def chat_completion(
        self,
        *,
        config: OpenAIChatConfig,
        messages: list[ChatCompletionMessageParam],
        tools: list[ChatCompletionToolParam] | None = None,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        stream: bool = False,
    ) -> AsyncStream[ChatCompletionChunk] | ChatCompletion:
        try:
            return await self._client.chat.completions.create(
                messages=messages,
                model=config.model,
                frequency_penalty=config.frequency_penalty
                if not_missing(config.frequency_penalty)
                else NOT_GIVEN,
                max_tokens=config.max_tokens if not_missing(config.max_tokens) else NOT_GIVEN,
                n=1,
                response_format=cast(ResponseFormat, config.response_format)
                if not_missing(config.response_format)
                else NOT_GIVEN,
                seed=config.seed if not_missing(config.seed) else NOT_GIVEN,
                stream=stream,
                temperature=config.temperature,
                tools=tools or NOT_GIVEN,
                tool_choice=tool_choice if tools else NOT_GIVEN,
                parallel_tool_calls=True if tools else NOT_GIVEN,
                top_p=config.top_p if not_missing(config.top_p) else NOT_GIVEN,
                timeout=config.timeout if not_missing(config.timeout) else NOT_GIVEN,
                stream_options={"include_usage": True} if stream else NOT_GIVEN,
            )

        except OpenAIRateLimitError as exc:  # retry on rate limit after delay
            if delay := exc.response.headers.get("Retry-After"):
                raise RateLimitError(
                    retry_after=delay + uniform(0.0, 0.3)  # nosec: B311 # add small random delay
                ) from exc

            else:
                raise exc

    async def embedding(
        self,
        config: OpenAIEmbeddingConfig,
        inputs: Sequence[str],
    ) -> list[list[float]]:
        return list(
            chain(
                *await gather(
                    *[
                        self._create_text_embedding(
                            texts=inputs[index : index + config.batch_size],
                            model=config.model,
                            dimensions=config.dimensions
                            if not_missing(config.dimensions)
                            else NOT_GIVEN,
                            encoding_format=cast(Literal["float", "base64"], config.encoding_format)
                            if not_missing(config.encoding_format)
                            else NOT_GIVEN,
                            timeout=config.timeout if not_missing(config.timeout) else NOT_GIVEN,
                        )
                        for index in range(0, len(inputs), config.batch_size)
                    ]
                )
            )
        )

    async def _create_text_embedding(
        self,
        texts: Sequence[str],
        model: str,
        dimensions: int | NotGiven,
        encoding_format: Literal["float", "base64"] | NotGiven,
        timeout: float | NotGiven,
    ) -> list[list[float]]:
        while True:
            try:
                response: CreateEmbeddingResponse = await self._client.embeddings.create(
                    input=list(texts),
                    model=model,
                    dimensions=dimensions,
                    encoding_format=encoding_format,
                    timeout=timeout,
                )
                return [element.embedding for element in response.data]

            except OpenAIRateLimitError as exc:  # always retry on rate limit after delay
                if delay := exc.response.headers.get("Retry-After"):
                    raise RateLimitError(
                        retry_after=delay + uniform(0.0, 0.3)  # nosec: B311 # add small random delay
                    ) from exc

                else:
                    raise exc

    async def moderation_check(
        self,
        text: str,
    ) -> Moderation:
        response: ModerationCreateResponse = await self._client.moderations.create(
            input=text,
        )
        return response.results[0]  # TODO: check API about multiple results

    async def generate_image(
        self,
        config: OpenAIImageGenerationConfig,
        instruction: str,
    ) -> Image:
        response: ImagesResponse = await self._client.images.generate(
            model=config.model,
            n=1,
            prompt=instruction,
            quality=config.quality,
            size=config.size,
            style=config.style,
            timeout=config.timeout if not_missing(config.timeout) else NOT_GIVEN,
            response_format=config.response_format,
        )
        return response.data[0]

    async def dispose(self) -> None:
        await self._client.close()



from typing import Literal

from draive.parameters import DataModel

__all__ = [
    "ImageBase64Content",
    "ImageContent",
    "ImageURLContent",
]


class ImageURLContent(DataModel):
    mime_type: Literal["image/jpeg", "image/png", "image/gif"] | None = None
    image_url: str
    image_description: str | None = None
    meta: dict[str, str | float | int | bool | None] | None = None


class ImageBase64Content(DataModel):
    mime_type: Literal["image/jpeg", "image/png", "image/gif"] | None = None
    image_base64: str
    image_description: str | None = None
    meta: dict[str, str | float | int | bool | None] | None = None


ImageContent = ImageURLContent | ImageBase64Content



from collections.abc import AsyncGenerator, Sequence
from typing import Any, Literal, cast, overload

from anthropic.types import (
    ImageBlockParam,
    Message,
    MessageParam,
    TextBlock,
    TextBlockParam,
    ToolParam,
    ToolUseBlock,
)

from draive.anthropic.client import AnthropicClient
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.errors import AnthropicException
from draive.instructions import Instruction
from draive.lmm import LMMToolSelection, ToolSpecification
from draive.metrics import ArgumentsTrace, ResultTrace, TokenUsage
from draive.parameters import DataModel
from draive.scope import ctx
from draive.types import (
    AudioBase64Content,
    AudioURLContent,
    ImageBase64Content,
    ImageURLContent,
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    MultimodalContent,
    MultimodalContentElement,
    TextContent,
    VideoBase64Content,
    VideoURLContent,
)

__all__ = [
    "anthropic_lmm_invocation",
]


@overload
async def anthropic_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def anthropic_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def anthropic_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def anthropic_lmm_invocation(  # noqa: PLR0913
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    with ctx.nested(
        "anthropic_lmm_invocation",
        metrics=[
            ArgumentsTrace.of(
                instruction=instruction,
                context=context,
                tools=tools,
                tool_selection=tool_selection,
                output=output,
                stream=stream,
                **extra,
            ),
        ],
    ):
        ctx.log_debug("Requested Anthropic lmm")
        client: AnthropicClient = ctx.dependency(AnthropicClient)
        config: AnthropicConfig = ctx.state(AnthropicConfig).updated(**extra)
        ctx.record(config)

        messages: list[MessageParam] = [
            _convert_context_element(element=element) for element in context
        ]

        if stream:
            return ctx.stream(
                _completion_stream(
                    client=client,
                    config=config,
                    instruction=Instruction.of(instruction).format(),
                    messages=messages,
                    tools=tools,
                    tool_selection=tool_selection,
                ),
            )

        else:
            return await _completion(
                client=client,
                config=config,
                instruction=Instruction.of(instruction).format(),
                messages=messages,
                tools=tools,
                tool_selection=tool_selection,
            )


def _convert_content_element(
    element: MultimodalContentElement,
) -> TextBlockParam | ImageBlockParam:
    match element:
        case TextContent() as text:
            return {
                "type": "text",
                "text": text.text,
            }

        case ImageURLContent() as image:
            # TODO: we could download the media to have data instead
            raise ValueError("Unsupported message content", element)

        case ImageBase64Content() as image:
            return {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image.mime_type or "image/png",
                    "data": image.image_base64,
                },
            }

        case AudioURLContent():
            raise ValueError("Unsupported message content", element)

        case AudioBase64Content():
            raise ValueError("Unsupported message content", element)

        case VideoURLContent():
            raise ValueError("Unsupported message content", element)

        case VideoBase64Content():
            raise ValueError("Unsupported message content", element)

        case DataModel() as data:
            return {
                "type": "text",
                "text": data.as_json(),
            }


def _convert_context_element(
    element: LMMContextElement,
) -> MessageParam:
    match element:
        case LMMInput() as input:
            return {
                "role": "user",
                "content": [
                    _convert_content_element(element=element) for element in input.content.parts
                ],
            }

        case LMMCompletion() as completion:
            # TODO: Anthropic models generating media?
            return {
                "role": "assistant",
                "content": completion.content.as_string(),
            }

        case LMMToolRequests() as tool_requests:
            return {
                "role": "assistant",
                "content": [
                    {
                        "id": request.identifier,
                        "type": "tool_use",
                        "name": request.tool,
                        "input": request.arguments,
                    }
                    for request in tool_requests.requests
                ],
            }

        case LMMToolResponse() as tool_response:
            return {
                "role": "user",
                "content": [
                    {
                        "tool_use_id": tool_response.identifier,
                        "type": "tool_result",
                        "is_error": tool_response.error,
                        "content": [
                            _convert_content_element(element=part)
                            for part in tool_response.content.parts
                        ],
                    }
                ],
            }


async def _completion(  # noqa: PLR0913, PLR0912, C901
    *,
    client: AnthropicClient,
    config: AnthropicConfig,
    instruction: str,
    messages: list[MessageParam],
    tools: Sequence[ToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> LMMOutput:
    completion: Message
    match tool_selection:
        case "auto":
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice="auto",
            )

        case "none":
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice="none",
            )

        case "required":
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice="any",
            )

        case tool:
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice={
                    "type": "tool",
                    "name": tool["function"]["name"],
                },
            )

    ctx.record(
        TokenUsage.for_model(
            config.model,
            input_tokens=completion.usage.input_tokens or 0,
            output_tokens=completion.usage.output_tokens or 0,
        ),
    )

    message_parts: list[TextBlock] = []
    tool_calls: list[ToolUseBlock] = []
    for part in completion.content:
        match part:
            case TextBlock() as text:
                message_parts.append(text)

            case ToolUseBlock() as call:
                tool_calls.append(call)

    match completion.stop_reason:
        case "tool_use":
            if (tool_calls := tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                raise AnthropicException("Invalid Anthropic completion", completion)

        case "end_turn":
            if (tool_calls := tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                ctx.record(ResultTrace.of(message_parts))
                return LMMCompletion.of(
                    MultimodalContent.of(*[TextContent(text=part.text) for part in message_parts])
                )

        case other:
            raise AnthropicException(f"Unexpected finish reason: {other}")


async def _completion_stream(  # noqa: PLR0913
    *,
    client: AnthropicClient,
    config: AnthropicConfig,
    instruction: str,
    messages: list[MessageParam],
    tools: Sequence[ToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> AsyncGenerator[LMMOutputStreamChunk, None]:
    ctx.log_debug("Anthropic streaming api is not supported yet, using regular response...")
    output: LMMOutput = await _completion(
        client=client,
        config=config,
        instruction=instruction,
        messages=messages,
        tools=tools,
        tool_selection=tool_selection,
    )

    match output:
        case LMMCompletion() as completion:
            yield LMMCompletionChunk.of(completion.content)

        case other:
            yield other



from asyncio import (
    AbstractEventLoop,
    Task,
    TaskGroup,
    current_task,
    get_running_loop,
    shield,
)
from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    Callable,
    Coroutine,
    Generator,
    Iterable,
)
from concurrent.futures import Executor
from contextvars import ContextVar, Token
from logging import Logger, getLogger
from types import TracebackType
from typing import Any, final

from draive.metrics import (
    ExceptionTrace,
    Metric,
    MetricsTrace,
    MetricsTraceReporter,
    metrics_log_reporter,
)
from draive.parameters import ParametrizedData
from draive.scope.dependencies import ScopeDependencies, ScopeDependency
from draive.scope.errors import MissingScopeContext
from draive.scope.state import ScopeState
from draive.utils import MISSING, AsyncStream, Missing, asynchronous, getenv_bool, mimic_function

__all__ = [
    "ctx",
]

_TaskGroup_Var = ContextVar[TaskGroup]("_TaskGroup_Var")
_MetricsScope_Var = ContextVar[MetricsTrace]("_MetricsScope_Var")
_StateScope_Var = ContextVar[ScopeState]("_ScopeState_Var")
_DependenciesScope_Var = ContextVar[ScopeDependencies]("_DependenciesScope_Var")


class _RootContext:
    def __init__(
        self,
        task_group: TaskGroup,
        dependencies: ScopeDependencies,
        state: ScopeState,
        metrics: MetricsTrace,
        trace_reporting: MetricsTraceReporter | None,
    ) -> None:
        self._task_group: TaskGroup = task_group
        self._task_group_token: Token[TaskGroup] | None = None
        self._dependencies: ScopeDependencies = dependencies
        self._dependencies_token: Token[ScopeDependencies] | None = None
        self._state: ScopeState = state
        self._state_token: Token[ScopeState] | None = None
        self._metrics: MetricsTrace = metrics
        self._metrics_token: Token[MetricsTrace] | None = None
        self._report_trace: MetricsTraceReporter | None = trace_reporting

    async def __aenter__(self) -> None:
        # start the task group first
        assert self._task_group_token is None, "Reentrance is not allowed"  # nosec: B101
        self._task_group_token = _TaskGroup_Var.set(self._task_group)
        await self._task_group.__aenter__()
        # prepare state scope
        assert self._state_token is None, "Reentrance is not allowed"  # nosec: B101
        self._state_token = _StateScope_Var.set(self._state)
        # then initialize dependencies
        assert self._dependencies_token is None, "Reentrance is not allowed"  # nosec: B101
        self._dependencies_token = _DependenciesScope_Var.set(self._dependencies)
        # finally begin metrics capture
        assert self._metrics_token is None, "Reentrance is not allowed"  # nosec: B101
        self._metrics.enter()
        self._metrics_token = _MetricsScope_Var.set(self._metrics)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._task_group_token is not None, "Can't exit scope without entering"  # nosec: B101
        _TaskGroup_Var.reset(self._task_group_token)
        # finish task group - wait for completion
        try:
            await self._task_group.__aexit__(
                et=exc_type,
                exc=exc_val,
                tb=exc_tb,
            )

        except BaseException as exc:
            # record task group exceptions
            self._metrics.record(ExceptionTrace.of(exc_val or exc))

        else:
            # or context exception
            if (exception := exc_val) and exc_type is not GeneratorExit:
                self._metrics.record(ExceptionTrace.of(exception))

        finally:
            # then end metrics capture
            assert self._metrics_token is not None, "Can't exit scope without entering"  # nosec: B101
            _MetricsScope_Var.reset(self._metrics_token)
            self._metrics.exit()
            assert (  # nosec: B101
                self._metrics.is_finished
            ), "Unbalanced metrics trace enter/exit calls, possibly an unstructured task running"
            # report metrics trace
            if report_trace := self._report_trace:
                # it still have access to the dependencies and state
                try:  # catch all exceptions - we don't want to blow up on metrics
                    await shield(
                        report_trace(
                            trace_id=self._metrics.trace_id,
                            logger=self._metrics._logger,  # pyright: ignore[reportPrivateUsage]
                            report=self._metrics.report(),
                        )
                    )

                except Exception as exc:
                    self._metrics.log_error(
                        "Failed to finish metrics trace report",
                        exception=exc,
                    )
                    return None

            # cleanup dependencies next
            assert self._dependencies_token is not None, "Can't exit scope without entering"  # nosec: B101
            _DependenciesScope_Var.reset(self._dependencies_token)
            # finally reset state
            assert self._state_token is not None, "Can't exit scope without entering"  # nosec: B101
            _StateScope_Var.reset(self._state_token)


class _PartialContext:
    def __init__(
        self,
        task_group: TaskGroup | None = None,
        metrics: MetricsTrace | None = None,
        state: ScopeState | None = None,
    ) -> None:
        self._task_group: TaskGroup | None = task_group
        self._task_group_token: Token[TaskGroup] | None = None
        self._metrics: MetricsTrace | None = metrics
        self._metrics_token: Token[MetricsTrace] | None = None
        self._state: ScopeState | None = state
        self._state_token: Token[ScopeState] | None = None

    def __enter__(self) -> None:
        if metrics := self._metrics:
            assert self._metrics_token is None, "Reentrance is not allowed"  # nosec: B101
            metrics.enter()
            self._metrics_token = _MetricsScope_Var.set(metrics)
        if state := self._state:
            assert self._state_token is None, "Reentrance is not allowed"  # nosec: B101
            self._state_token = _StateScope_Var.set(state)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if (token := self._metrics_token) and (metrics := self._metrics):
            _MetricsScope_Var.reset(self._metrics_token)
            if (exception := exc_val) and exc_type is not GeneratorExit:
                metrics.record(ExceptionTrace.of(exception))

            metrics.exit()

        if token := self._state_token:
            _StateScope_Var.reset(token)

    async def __aenter__(self) -> None:
        if task_group := self._task_group:
            assert self._task_group_token is None, "Reentrance is not allowed"  # nosec: B101
            self._task_group_token = _TaskGroup_Var.set(task_group)
            await task_group.__aenter__()

        self.__enter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        try:
            if task_group := self._task_group:
                assert self._task_group_token is not None, "Can't exit scope without entering"  # nosec: B101
                _TaskGroup_Var.reset(self._task_group_token)
                await task_group.__aexit__(
                    et=exc_type,
                    exc=exc_val,
                    tb=exc_tb,
                )

        except BaseException as exc:
            self.__exit__(
                exc_type=type(exc),
                exc_val=exc,
                exc_tb=exc.__traceback__,
            )

        else:
            self.__exit__(
                exc_type=exc_type,
                exc_val=exc_val,
                exc_tb=exc_tb,
            )


@final
class ctx:
    @staticmethod
    def new(  # noqa: PLR0913
        label: str | None = None,
        /,
        *,
        dependencies: ScopeDependencies
        | Iterable[type[ScopeDependency] | ScopeDependency]
        | None = None,
        state: ScopeState | Iterable[ParametrizedData] | None = None,
        metrics: Iterable[Metric] | None = None,
        logger: Logger | None = None,
        trace_reporting: MetricsTraceReporter | None = None,
    ) -> _RootContext:
        root_dependencies: ScopeDependencies
        if dependencies is None:
            root_dependencies = ScopeDependencies()
        elif isinstance(dependencies, ScopeDependencies):
            root_dependencies = dependencies
        else:
            root_dependencies = ScopeDependencies(*dependencies)

        root_state: ScopeState
        if state is None:
            root_state = ScopeState()
        elif isinstance(state, ScopeState):
            root_state = state
        else:
            root_state = ScopeState(*state)

        root_logger: Logger = logger or getLogger(name=label)

        trace_reporter: MetricsTraceReporter | None
        if trace_reporting:
            trace_reporter = trace_reporting
        elif getenv_bool("DEBUG_LOGGING", __debug__):
            trace_reporter = metrics_log_reporter(
                list_items_limit=-4,
                item_character_limit=64,
            )
        else:
            trace_reporter = None

        return _RootContext(
            task_group=TaskGroup(),
            dependencies=root_dependencies,
            state=root_state,
            metrics=MetricsTrace(
                label=label,
                logger=root_logger,
                parent=None,
                metrics=metrics,
            ),
            trace_reporting=trace_reporter,
        )

    @staticmethod
    def wrap[**Args, Result](  # noqa: PLR0913
        label: str | None = None,
        /,
        *,
        dependencies: ScopeDependencies
        | Iterable[type[ScopeDependency] | ScopeDependency]
        | None = None,
        state: ScopeState | Iterable[ParametrizedData] | None = None,
        metrics: Iterable[Metric] | None = None,
        logger: Logger | None = None,
        trace_reporting: MetricsTraceReporter | None = None,
    ) -> Callable[
        [Callable[Args, Coroutine[None, None, Result]]],
        Callable[Args, Coroutine[None, None, Result]],
    ]:
        root_dependencies: ScopeDependencies
        if dependencies is None:
            root_dependencies = ScopeDependencies()
        elif isinstance(dependencies, ScopeDependencies):
            root_dependencies = dependencies
        else:
            root_dependencies = ScopeDependencies(*dependencies)

        root_state: ScopeState
        if state is None:
            root_state = ScopeState()
        elif isinstance(state, ScopeState):
            root_state = state
        else:
            root_state = ScopeState(*state)

        trace_reporter: MetricsTraceReporter | None
        if trace_reporting:
            trace_reporter = trace_reporting
        elif getenv_bool("DEBUG_LOGGING", __debug__):
            trace_reporter = metrics_log_reporter(
                list_items_limit=-4,
                item_character_limit=64,
            )
        else:
            trace_reporter = None

        def wrapper(
            function: Callable[Args, Coroutine[None, None, Result]],
            /,
        ) -> Callable[Args, Coroutine[None, None, Result]]:
            @mimic_function(function)
            async def wrapped(*args: Args.args, **kwargs: Args.kwargs) -> Result:
                async with ctx.new(
                    label,
                    dependencies=root_dependencies,
                    state=root_state,
                    metrics=metrics,
                    logger=logger,
                    trace_reporting=trace_reporter,
                ):
                    return await function(*args, **kwargs)

            return wrapped

        return wrapper

    @staticmethod
    def update[**Args, Result](
        *state: ParametrizedData,
    ) -> Callable[
        [Callable[Args, Coroutine[None, None, Result]]],
        Callable[Args, Coroutine[None, None, Result]],
    ]:
        def wrapper(
            function: Callable[Args, Coroutine[None, None, Result]],
            /,
        ) -> Callable[Args, Coroutine[None, None, Result]]:
            @mimic_function(function)
            async def wrapped(*args: Args.args, **kwargs: Args.kwargs) -> Result:
                with ctx.updated(*state):
                    return await function(*args, **kwargs)

            return wrapped

        return wrapper

    @staticmethod
    def _current_task_group() -> TaskGroup:
        try:
            return _TaskGroup_Var.get()

        except LookupError as exc:
            raise MissingScopeContext("TaskGroup requested but not defined!") from exc

    @staticmethod
    def _current_metrics() -> MetricsTrace:
        try:
            return _MetricsScope_Var.get()

        except LookupError as exc:
            raise MissingScopeContext("MetricsScope requested but not defined!") from exc

    @staticmethod
    def _current_dependencies() -> ScopeDependencies:
        try:
            return _DependenciesScope_Var.get()

        except LookupError as exc:
            raise MissingScopeContext("DependenciesScope requested but not defined!") from exc

    @staticmethod
    def _current_state() -> ScopeState:
        try:
            return _StateScope_Var.get()

        except LookupError as exc:
            raise MissingScopeContext("StateScope requested but not defined!") from exc

    @staticmethod
    def spawn_task[**Args, Result](
        function: Callable[Args, Coroutine[None, None, Result]],
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Task[Result]:
        nested_context: _PartialContext = ctx.nested(function.__name__)

        async def wrapped(*args: Args.args, **kwargs: Args.kwargs) -> Result:
            with nested_context:
                return await function(*args, **kwargs)

        return ctx._current_task_group().create_task(wrapped(*args, **kwargs))

    @staticmethod
    def spawn_subtask[**Args, Result](
        function: Callable[Args, Coroutine[None, None, Result]],
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Task[Result]:
        return ctx._current_task_group().create_task(function(*args, **kwargs))

    @staticmethod
    def cancel() -> None:
        if task := current_task():
            task.cancel()

        else:
            raise RuntimeError("Attempting to cancel context out of asyncio task")

    @staticmethod
    def nested(
        label: str,
        /,
        *,
        state: ScopeState | Iterable[ParametrizedData] | None = None,
        metrics: Iterable[Metric] | None = None,
    ) -> _PartialContext:
        nested_state: ScopeState | None
        if isinstance(state, ScopeState):
            nested_state = state
        else:
            nested_state = ctx._current_state().updated(state)

        return _PartialContext(
            task_group=TaskGroup(),
            metrics=ctx._current_metrics().nested(
                label=label,
                metrics=metrics,
            ),
            state=nested_state,
        )

    @staticmethod
    def stream[Element](
        source: AsyncGenerator[Element, None],
        /,
    ) -> AsyncIterable[Element]:
        metrics: MetricsTrace = ctx._current_metrics()
        metrics.enter()

        stream: AsyncStream[Element] = AsyncStream()

        async def consumer() -> None:
            try:
                async for element in source:
                    await stream.send(element)

            except BaseException as exc:
                stream.finish(exception=exc)

            else:
                stream.finish()

        task: Task[None] = ctx.spawn_subtask(consumer)

        def on_finish(task: Task[None]) -> None:
            metrics.exit()

        task.add_done_callback(on_finish)

        return stream

    @staticmethod
    def stream_sync[Element](
        source: Generator[Element, None],
        /,
        executor: Executor | None = None,
    ) -> AsyncIterable[Element]:
        metrics: MetricsTrace = ctx._current_metrics()
        metrics.enter()

        loop: AbstractEventLoop = get_running_loop()

        @asynchronous(loop=loop, executor=executor)
        def source_next() -> Element:
            try:
                return next(source)

            except GeneratorExit as exc:
                raise StopAsyncIteration() from exc

            except StopIteration as exc:
                raise StopAsyncIteration() from exc

        stream: AsyncStream[Element] = AsyncStream()

        async def consumer() -> None:
            try:
                while True:
                    try:
                        await stream.send(await source_next())

                    except StopAsyncIteration:
                        break

            except BaseException as exc:
                stream.finish(exception=exc)

            else:
                stream.finish()

        task: Task[None] = ctx.spawn_subtask(consumer)

        def on_finish(task: Task[None]) -> None:
            metrics.exit()

        task.add_done_callback(on_finish)

        return stream

    @staticmethod
    def updated(
        *state: ParametrizedData,
    ) -> _PartialContext:
        return _PartialContext(state=ctx._current_state().updated(state))

    @staticmethod
    def id() -> str:
        return ctx._current_metrics().trace_id

    @staticmethod
    def state[State_T: ParametrizedData](
        state: type[State_T],
        /,
        default: State_T | Missing = MISSING,
    ) -> State_T:
        return ctx._current_state().state(
            state,
            default=default,
        )

    @staticmethod
    def dependency[Dependency_T: ScopeDependency](
        dependency: type[Dependency_T],
        /,
    ) -> Dependency_T:
        return ctx._current_dependencies().dependency(dependency)

    @staticmethod
    def read[Metric_T: Metric](
        metric: type[Metric_T],
        /,
    ) -> Metric_T | None:
        return ctx._current_metrics().read(metric)

    @staticmethod
    def record(
        *metrics: Metric,
    ) -> None:
        try:
            ctx._current_metrics().record(*metrics)

        # ignoring metrics record when using out of metrics context
        # using default logger as fallback as we already know that we are missing metrics
        except MissingScopeContext as exc:
            logger: Logger = getLogger()
            logger.error("Attempting to record metrics outside of metrics context")
            logger.error(
                exc,
                exc_info=True,
            )

    @staticmethod
    def log_error(
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        try:
            ctx._current_metrics().log_error(
                message,
                *args,
                exception=exception,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            logger: Logger = getLogger()
            logger.error(
                message,
                *args,
            )
            if exception := exception:
                logger.error(
                    exception,
                    exc_info=True,
                )

    @staticmethod
    def log_warning(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            ctx._current_metrics().log_warning(
                message,
                *args,
                exception=exception,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            logger: Logger = getLogger()
            logger.warning(
                message,
                *args,
            )
            if exception := exception:
                logger.error(
                    exception,
                    exc_info=True,
                )

    @staticmethod
    def log_info(
        message: str,
        /,
        *args: Any,
    ) -> None:
        try:
            ctx._current_metrics().log_info(
                message,
                *args,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            getLogger().info(
                message,
                *args,
            )

    @staticmethod
    def log_debug(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            ctx._current_metrics().log_debug(
                message,
                *args,
                exception=exception,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            logger: Logger = getLogger()
            logger.debug(
                message,
                *args,
            )
            if exception := exception:
                logger.error(
                    exception,
                    exc_info=True,
                )



from typing import Any

from anthropic import Anthropic
from tokenizers import Tokenizer

from draive.utils import cache

__all__ = [
    "anthropic_tokenize_text",
]


def anthropic_tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    return _tokenizer().encode(text).ids  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]


@cache(limit=1)
def _tokenizer() -> Tokenizer:
    return Anthropic().get_tokenizer()



from typing import Any, Protocol, runtime_checkable

__all__ = [
    "TextTokenizer",
]


@runtime_checkable
class TextTokenizer(Protocol):
    def __call__(
        self,
        text: str,
        **extra: Any,
    ) -> list[int]: ...



import builtins
import types
import typing
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Callable, Mapping, Sequence
from copy import copy
from typing import Any, final, get_args, get_origin, overload

from draive.utils import MISSING, Missing, freeze, not_missing

__all__ = [
    "ParameterPath",
]


class ParameterPathComponent(ABC):
    @abstractmethod
    def path_str(
        self,
        current: str | None = None,
    ) -> str: ...

    @abstractmethod
    def resolve(
        self,
        subject: Any,
        /,
    ) -> Any: ...

    @abstractmethod
    def assign(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any: ...


@final
class ParameterPathAttributeComponent(ParameterPathComponent):
    def __init__[Root, Parameter](
        self,
        root: type[Root],
        parameter: type[Parameter],
        *,
        attribute: str,
    ) -> None:
        root_origin: Any = get_origin(root) or root
        parameter_origin: Any = get_origin(parameter) or parameter

        def resolve(
            subject: Root,
            /,
        ) -> Parameter:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"ParameterPathComponent used on unexpected root of "
                f"'{type(root)}' instead of '{root}' for '{attribute}'"
            )

            resolved: Any = getattr(subject, attribute)

            assert isinstance(resolved, parameter_origin), (  # nosec: B101
                f"ParameterPathComponent pointing to unexpected value of "
                f"'{type(resolved)}' instead of '{parameter}' for '{attribute}'"
            )
            return resolved

        def assign(
            subject: Root,
            /,
            value: Parameter,
        ) -> Root:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"ParameterPathComponent used on unexpected root of "
                f"'{type(root)}' instead of '{root}' for '{attribute}'"
            )
            assert isinstance(value, parameter_origin), (  # nosec: B101
                f"ParameterPathComponent assigning to unexpected value of "
                f"'{type(value)}' instead of '{parameter}' for '{attribute}'"
            )

            updated: Root
            if hasattr(subject, "updating_parameter"):  # can't check full type here
                updated = subject.updating_parameter(attribute, value=value)  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportAttributeAccessIssue]

            else:
                updated = copy(subject)
                setattr(updated, attribute, value)

            return updated  # pyright: ignore[reportUnknownVariableType]

        self._resolve: Callable[[Root], Parameter] = resolve
        self._assign: Callable[[Root, Parameter], Root] = assign
        self._attribute: str = attribute

        freeze(self)

    def path_str(
        self,
        current: str | None = None,
    ) -> str:
        if current:
            return f"{current}.{self._attribute}"
        else:
            return self._attribute

    def resolve(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._resolve(subject)

    def assign(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any:
        return self._assign(subject, value)


@final
class ParameterPathItemComponent(ParameterPathComponent):
    def __init__[Root, Parameter](
        self,
        root: type[Root],
        parameter: type[Parameter],
        *,
        item: Any,
    ) -> None:
        root_origin: Any = get_origin(root) or root
        parameter_origin: Any = get_origin(parameter) or parameter

        def resolve(
            subject: Root,
            /,
        ) -> Parameter:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"ParameterPathComponent used on unexpected root of "
                f"'{type(root)}' instead of '{root}' for '{item}'"
            )

            resolved: Any = subject.__getitem__(item)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]

            assert isinstance(resolved, parameter_origin), (  # nosec: B101
                f"ParameterPathComponent pointing to unexpected value of "
                f"'{type(resolved)}' instead of '{parameter}' for '{item}'"
            )
            return resolved

        def assign(
            subject: Root,
            /,
            value: Parameter,
        ) -> Root:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"ParameterPathComponent used on unexpected root of "
                f"'{type(root)}' instead of '{root}' for '{item}'"
            )
            assert isinstance(value, parameter_origin), (  # nosec: B101
                f"ParameterPathComponent assigning to unexpected value of "
                f"'{type(value)}' instead of '{parameter}' for '{item}'"
            )

            if hasattr(subject, "__setitem__"):  # can't check full type here
                updated: Root = copy(subject)
                updated.__setitem__(item, value)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                return updated

            elif isinstance(subject, Mapping):
                temp_dict: dict[Any, Any] = dict(subject)  # pyright: ignore[reportUnknownArgumentType]
                temp_dict[item] = value
                return subject.__class__(temp_dict)  # pyright: ignore[reportCallIssue, reportUnknownVariableType, reportUnknownMemberType]

            elif isinstance(subject, Sequence):
                temp_list: list[Any] = list(subject)  # pyright: ignore[reportUnknownArgumentType]
                temp_list[item] = value
                return subject.__class__(temp_list)  # pyright: ignore[reportCallIssue, reportUnknownVariableType, reportUnknownMemberType]

            else:
                raise RuntimeError(f"Unsupported item assignment - {type(value)} in {type(root)}")

        self._resolve: Callable[[Root], Parameter] = resolve
        self._assign: Callable[[Root, Parameter], Root] = assign
        self._item: Any = item

    def path_str(
        self,
        current: str | None = None,
    ) -> str:
        if current:
            return f"{current}[{self._item}]"
        else:
            return f"[{self._item}]"

    def resolve(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._resolve(subject)

    def assign(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any:
        return self._assign(subject, value)


@final
class ParameterPath[Root, Parameter]:
    @overload
    def __init__(
        self,
        root: type[Root],
        parameter: type[Root],
    ) -> None: ...

    @overload
    def __init__(
        self,
        root: type[Root],
        parameter: type[Parameter],
        *path: ParameterPathComponent,
    ) -> None: ...

    def __init__(
        self,
        root: type[Root],
        parameter: type[Parameter],
        *components: ParameterPathComponent,
    ) -> None:
        assert components or root == parameter  # nosec: B101
        self.__root__: type[Root] = root
        self.__parameter__: type[Parameter] = parameter
        self.__components__: tuple[ParameterPathComponent, ...] = components

        freeze(self)

    def components(self) -> tuple[str, ...]:
        return tuple(component.path_str() for component in self.__components__)

    def __str__(self) -> str:
        path: str = ""
        for component in self.__components__:
            path = component.path_str(path)
        return path

    def __repr__(self) -> str:
        path: str = self.__root__.__qualname__
        for component in self.__components__:
            path = component.path_str(path)
        return path

    def __getattr__(
        self,
        name: str,
    ) -> Any:
        try:
            return object.__getattribute__(self, name)

        except (AttributeError, KeyError):
            pass  # continue

        assert not name.startswith(  # nosec: B101
            "_"
        ), f"Accessing private/special parameter paths ({name}) is forbidden"

        try:
            annotation: Any = self.__parameter__.__annotations__[name]

        except (AttributeError, KeyError) as exc:
            raise AttributeError(name) from exc

        return ParameterPath[Root, Any](
            self.__root__,
            annotation,
            *[
                *self.__components__,
                ParameterPathAttributeComponent(
                    root=self.__parameter__,
                    parameter=annotation,
                    attribute=name,
                ),
            ],
        )

    def __getitem__(
        self,
        key: str | int,
    ) -> Any:
        annotation: Any
        match get_origin(self.__parameter__) or self.__parameter__:
            case (
                builtins.list  # pyright: ignore[reportUnknownMemberType]
                | typing.List  # pyright: ignore[reportUnknownMemberType]  # noqa: UP006
            ):
                match get_args(self.__parameter__):
                    case (element_annotation,) if isinstance(key, int):
                        annotation = element_annotation

                    case other:
                        raise TypeError("Unsupported list type annotation", other)
            case (
                builtins.tuple  # pyright: ignore[reportUnknownMemberType]
                | typing.Tuple  # pyright: ignore[reportUnknownMemberType]  # noqa: UP006
            ):
                match get_args(self.__parameter__):
                    case (element_annotation, builtins.Ellipsis | types.EllipsisType):
                        annotation = element_annotation

                    case other if isinstance(key, int):
                        annotation = other[key]

                    case other:
                        raise TypeError("Unsupported type annotation", other)
            case (
                builtins.dict  # pyright: ignore[reportUnknownMemberType]
                | typing.Dict  # pyright: ignore[reportUnknownMemberType]  # noqa: UP006
            ):
                match get_args(self.__parameter__):
                    case (builtins.str, element_annotation) if isinstance(key, str):
                        annotation = element_annotation

                    case (builtins.int, element_annotation) if isinstance(key, int):
                        annotation = element_annotation

                    case other:  # pyright: ignore[reportUnnecessaryComparison]
                        raise TypeError("Unsupported dict type annotation", other)

            case other:
                raise TypeError("Unsupported type annotation", other)

        return ParameterPath[Root, Any](
            self.__root__,
            annotation,
            *[
                *self.__components__,
                ParameterPathItemComponent(
                    root=self.__parameter__,
                    parameter=annotation,
                    item=key,
                ),
            ],
        )

    @overload
    def __call__(
        self,
        root: Root,
        /,
    ) -> Parameter: ...

    @overload
    def __call__(
        self,
        root: Root,
        /,
        updated: Parameter,
    ) -> Root: ...

    def __call__(
        self,
        root: Root,
        /,
        updated: Parameter | Missing = MISSING,
    ) -> Root | Parameter:
        assert isinstance(root, get_origin(self.__root__) or self.__root__), (  # nosec: B101
            f"ParameterPath '{self.__repr__()}' used on unexpected root of "
            f"'{type(root)}' instead of '{self.__root__}'"
        )

        if not_missing(updated):
            assert isinstance(updated, get_origin(self.__parameter__) or self.__parameter__), (  # nosec: B101
                f"ParameterPath '{self.__repr__()}' assigning to unexpected value of "
                f"'{type(updated)}' instead of '{self.__parameter__}'"
            )

            resolved: Any = root
            updates_stack: deque[tuple[Any, ParameterPathComponent]] = deque()
            for component in self.__components__:
                updates_stack.append((resolved, component))
                resolved = component.resolve(resolved)

            updated_value: Any = updated
            while updates_stack:
                subject, component = updates_stack.pop()
                updated_value = component.assign(
                    subject,
                    value=updated_value,
                )

            return updated_value

        else:
            resolved: Any = root
            for component in self.__components__:
                resolved = component.resolve(resolved)

            assert isinstance(resolved, get_origin(self.__parameter__) or self.__parameter__), (  # nosec: B101
                f"ParameterPath '{self.__repr__()}' pointing to unexpected value of "
                f"'{type(resolved)}' instead of '{self.__parameter__}'"
            )

            return resolved



from typing import Any

from draive.instructions.state import InstructionsRepository
from draive.instructions.types import Instruction
from draive.scope import ctx

__all__ = [
    "instruction",
]


async def instruction(
    key: str,
    /,
    *,
    default: Instruction | str | None = None,
    **extra: Any,
) -> Instruction:
    return (
        await ctx.state(InstructionsRepository)
        .updated(**extra)
        .instruction(
            key,
            default=default,
        )
    )



from typing import Any

from tiktoken import Encoding, encoding_for_model

from draive.openai.config import OpenAIChatConfig
from draive.scope import ctx
from draive.utils import cache

__all__ = [
    "openai_tokenize_text",
]


def openai_tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    return _encoding(model_name=ctx.state(OpenAIChatConfig).model).encode(text=text)


@cache(limit=4)
def _encoding(model_name: str) -> Encoding:
    return encoding_for_model(model_name=model_name)



from collections.abc import Iterable
from logging import Logger, getLogger
from time import monotonic
from typing import Any, Self, cast, final
from uuid import uuid4

from draive.metrics.metric import Metric
from draive.metrics.reporter import MetricsTraceReport

__all__ = [
    "MetricsTrace",
]


@final  # unstructured background tasks spawning may result in corrupted data
class MetricsTrace:
    def __init__(
        self,
        *,
        label: str | None,
        logger: Logger | None,
        parent: Self | None,
        metrics: Iterable[Metric] | None,
    ) -> None:
        self._start: float = monotonic()
        self._end: float | None = None
        self._runners: int = 0
        self._trace_id: str = parent._trace_id if parent else uuid4().hex
        self._label: str = label or "metrics"
        self._parent: Self | None = parent
        self._logger: Logger = logger or (parent._logger if parent else getLogger(name=self._label))
        self._metrics: dict[type[Metric], Metric] = {
            type(metric): metric for metric in metrics or []
        }
        self._nested_traces: list[MetricsTrace] = []
        self.log_info("started...")

    # - STATE -

    @property
    def trace_id(self) -> str:
        return self._trace_id

    @property
    def is_root(self) -> bool:
        return self._parent is None

    @property
    def is_finished(self) -> bool:
        return self._end is not None

    def nested(
        self,
        label: str,
        *,
        metrics: Iterable[Metric] | None = None,
    ) -> Self:
        if self.is_finished:
            raise ValueError("Attempting to use already finished metrics trace")
        self.enter()
        nested: Self = self.__class__(
            label=label,
            logger=self._logger,
            parent=self,
            metrics=metrics,
        )
        self._nested_traces.append(nested)
        return nested

    # - METRICS -

    def record(
        self,
        *metrics: Metric,
    ) -> None:
        if self.is_finished:  # ignore record when already finished
            return self.log_error("Attempting to use already finished metrics trace, ignoring...")

        for metric in metrics:
            metric_type: type[Metric] = type(metric)
            try:  # catch exceptions - we don't wan't to blow up on metrics
                if current := self._metrics.get(metric_type):
                    if hasattr(current, "__add__"):
                        self._metrics[metric_type] = current + metric  # pyright: ignore[reportOperatorIssue]
                    else:
                        raise NotImplementedError(f"{metric_type.__qualname__} can't be combined!")

                else:
                    self._metrics[metric_type] = metric

            except Exception as exc:
                self.log_error(
                    "Failed to record %s metric",
                    metric_type.__qualname__,
                    exception=exc,
                )

    def read[Metric_T: Metric](
        self,
        metric: type[Metric_T],
        /,
    ) -> Metric_T | None:
        try:  # catch all exceptions - we don't wan't to blow up on metrics
            return cast(Metric_T, self._metrics.get(metric))

        except Exception as exc:
            self.log_error(
                "Failed to read %s metric",
                metric.__qualname__,
                exception=exc,
            )
            return None

    # - LOGS -

    def log_error(
        self,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        self._logger.error(
            f"[%s] {message}",
            self,
            *args,
        )
        if exception := exception:
            self._logger.error(
                exception,
                exc_info=True,
            )

    def log_warning(
        self,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        self._logger.warning(
            f"[%s] {message}",
            self,
            *args,
        )
        if exception := exception:
            self._logger.warning(
                exception,
                exc_info=True,
            )

    def log_info(
        self,
        message: str,
        /,
        *args: Any,
    ) -> None:
        self._logger.info(
            f"[%s] {message}",
            self,
            *args,
        )

    def log_debug(
        self,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        self._logger.debug(
            f"[%s] {message}",
            self,
            *args,
        )
        if exception := exception:
            self._logger.debug(
                exception,
                exc_info=True,
            )

    # - INTERNAL -

    def enter(self) -> None:
        assert not self.is_finished, "Attempting to use already finished metrics trace"  # nosec: B101
        self._runners += 1

    def exit(self) -> None:
        assert not self.is_finished, "Attempting to use already finished metrics trace"  # nosec: B101
        assert self._runners > 0, "Unbalanced metrics trace exit call"  # nosec: B101
        self._runners -= 1

        if self._runners > 0:
            return  # can't finish yet

        self._end = monotonic()

        self.log_info(
            "...finished after %.2fs",
            self._end - self._start,
        )

        if parent := self._parent:
            parent.exit()

    def report(self) -> MetricsTraceReport:
        return MetricsTraceReport(
            label=self._label,
            duration=(self._end or monotonic()) - self._start,
            metrics={key.__qualname__: value for key, value in self._metrics.items()},
            nested=[trace.report() for trace in self._nested_traces],
            combined=False,
        )

    def __str__(self) -> str:
        return f"{self._trace_id}|{self._label}"



from typing import Any, Self

from draive.utils import freeze

__all__ = [
    "ParameterValidationContext",
    "ParameterValidationError",
]


class ParameterValidationContext:
    def __init__(
        self,
        path: tuple[str, ...],
    ) -> None:
        self._path: tuple[str, ...] = path

        freeze(self)

    def __str__(self) -> str:
        return "".join(self._path)

    def __repr__(self) -> str:
        return "".join(self._path)

    def appending_path(
        self,
        path: str,
        /,
    ) -> Self:
        return self.__class__(
            path=(*self._path, path),
        )


class ParameterValidationError(Exception):
    @classmethod
    def missing(
        cls,
        context: ParameterValidationContext,
    ) -> Self:
        return cls("Validation error: Missing required parameter at: %s", context)

    @classmethod
    def invalid_type(
        cls,
        expected: Any,
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter at: %s - expected '%s' while received '%s'",
            context,
            getattr(expected, "__name__", str(expected)),
            getattr(type(received), "__name__", str(received)),  # pyright: ignore[reportUnknownArgumentType]
        )

    @classmethod
    def invalid_union_type(
        cls,
        expected: tuple[Any, ...],
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter at: %s - expected %s while received '%s'",
            context,
            " | ".join(getattr(t, "__name__", str(t)) for t in expected),
            received,
        )

    @classmethod
    def invalid_key(
        cls,
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter key at: %s - received '%s'",
            context,
            received,
        )

    @classmethod
    def invalid_value(
        cls,
        expected: Any,
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter value at: %s - expected '%s' while received '%s'",
            context,
            expected,
            received,
        )

    @classmethod
    def invalid(
        cls,
        exception: Exception,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter at: %s - %s",
            context,
            exception,
        )



import re
from collections.abc import Callable, Sequence
from re import Match, Pattern

__all__ = [
    "exhaustive_split_text",
    "exhaustive_regex_split_text",
]


def exhaustive_split_text(
    text: str,
    part_size: int,
    count_size: Callable[[str], int],
    separators: Sequence[str] | str | None = None,
    part_overlap_size: int | None = None,
) -> list[str]:
    # if the text is already small enough just use it
    if count_size(text) <= part_size:
        return [text]

    splitters: Sequence[str]

    match separators:
        case None:
            splitters = ["\n\n", "\n", " "]

        case str() as splitter:
            splitters = [splitter, " "]

        case [*separators]:
            splitters = separators

    # split using provided separators
    parts: list[str] = _split(
        text=text,
        splitters=splitters,
    )
    # then merge
    return _merge(
        parts=parts,
        part_size=part_size,
        count_size=count_size,
        part_overlap_size=part_overlap_size,
    )


def _split(
    text: str,
    splitters: Sequence[str],
) -> list[str]:
    result: list[str] = [text]

    for splitter in splitters:
        parts: list[str] = []
        for element in result:
            splitted: list[str] = element.split(splitter)
            for part in splitted[:-1]:
                parts.append(part + splitter)
            parts.append(splitted[-1])
        result = parts

    return result


def exhaustive_regex_split_text(
    text: str,
    part_size: int,
    count_size: Callable[[str], int],
    separators: Sequence[str | Pattern[str]] | str | Pattern[str] | None = None,
    part_overlap_size: int | None = None,
) -> list[str]:
    if count_size(text) <= part_size:
        return [text]

    splitters: Sequence[str | Pattern[str]]

    match separators:
        case None:
            splitters = ["(\n\n)", "(\n)", r"(\s)"]
        case str() | Pattern() as splitter:
            splitters = [splitter, r"(\s)"]
        case [*separators]:
            splitters = separators

    parts: list[str] = _regex_split(
        text=text,
        splitters=splitters,
    )
    return _merge(
        parts=parts,
        part_size=part_size,
        count_size=count_size,
        part_overlap_size=part_overlap_size,
    )


def _regex_split(
    text: str,
    splitters: Sequence[str | Pattern[str]],
) -> list[str]:
    result: list[str] = [text]

    for splitter in splitters:
        parts: list[str] = []
        for element in result:
            splitted: list[str] = re.split(
                f"({_remove_capturing_groups(splitter)})",
                element,
            )
            # Add the first part (always a non-splitter part)
            parts.append(splitted[0])
            # Combine splitter and following part
            parts.extend(
                [
                    splitter + following_part
                    for splitter, following_part in zip(
                        splitted[1::2],
                        splitted[2::2],
                        strict=False,
                    )
                ]
            )
        result = parts

    return result


def _remove_capturing_groups(pattern: str | Pattern[str]) -> str:
    if isinstance(pattern, Pattern):
        pattern = pattern.pattern

    def process_group(match: Match[str]) -> str:
        group_type, group_content = match.groups()
        special_constructs: set[str] = {"?:", "?=", "?!", "?<=", "?<!", "?>"}

        if group_type in special_constructs:
            return match.group(0)  # Return special constructs unchanged

        return group_content  # Remove parentheses for normal capturing groups

    # Regex to match all types of parentheses groups
    group_pattern = r"""\(  # Match opening parenthesis
                        (\??[=!<>:]?)  # Optional special characters for non-capturing groups
                        (  # Main content of the group
                            (?:
                                [^()\\]  # Any character that's not a parenthesis or backslash
                                |  # OR
                                \\.  # An escaped character
                                |  # OR
                                \((?:[^()\\]|\\.)*\)  # A nested group
                            )*
                        )
                    \)  # Match closing parenthesis
                    """

    processed_pattern: str = re.sub(
        group_pattern,
        process_group,
        pattern,
        flags=re.VERBOSE,
    )

    return processed_pattern


def _merge(
    parts: list[str],
    part_size: int,
    count_size: Callable[[str], int],
    part_overlap_size: int | None,
) -> list[str]:
    result: list[str] = []
    accumulator: str = ""
    overlap_accumulator: list[str] = []

    # iterate over splitted parts
    for part in parts:
        merged_part: str = accumulator + part
        # check if can add to previous part
        if count_size(merged_part) <= part_size:
            accumulator = merged_part
            overlap_accumulator.append(part)

        # check if current part is not too big on its own
        elif count_size(part) > part_size:
            raise ValueError("Failed to split text fitting required size.")

        # if we have overlap defined do overlap between last part and current (not fitting)
        elif part_overlap_size := part_overlap_size:
            if chunk := accumulator.strip():
                result.append(chunk)

            # clear accumulator - we will fill it now
            accumulator = ""
            for element in reversed(overlap_accumulator):
                merged_accumulator: str = element + accumulator

                if (
                    count_size(merged_accumulator) < part_overlap_size
                    and count_size(accumulator + part) <= part_size
                ):
                    overlap_accumulator = [element, *overlap_accumulator]
                    accumulator = merged_accumulator

                else:
                    break

            overlap_accumulator.append(part)
            accumulator = accumulator + part

        # otherwise start a new part out of current
        else:
            if chunk := accumulator.strip():
                result.append(chunk)

            accumulator = part
            overlap_accumulator = [part]

    # add leftover if any
    if chunk := accumulator.strip():
        result.append(chunk)

    return result



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.score import CommonScoreModel
from draive.generation import generate_model

__all__ = [
    "text_coverage_evaluator",
]


INSTRUCTION: str = """\
You will be given a reference text and a compared text based on the reference text.
Your task is to rate the compared text using only the Coverage metric, \
which is described in the Evaluation Criteria.
Please make sure you read and understand these instructions very carefully.
Keep this document open while reviewing, and refer to it as needed.

Evaluation Criteria:
Coverage (0.0-4.0) - the extent to which the compared text includes all \
the key points from the reference text.
A compared text with good coverage includes all the important information from \
the reference text without omitting critical points.
Annotators should penalize compared texts that miss significant content.

Rating Scale:
0.0: Very low coverage - the text misses most key points from the reference text.
1.0: Low coverage - the text includes some key points but omits several important ones.
2.0: Moderate coverage - the text covers most key points but misses a few important details.
3.0: Good coverage - the text includes nearly all key points with minor omissions.
4.0: Excellent coverage - the text comprehensively covers all key points from the reference text.

Evaluation Steps:
1. Read the reference text carefully and identify all key points and important information.
2. Read the compared text and compare it to the reference text. \
Check if the compared text includes all the key points and important information \
from the reference text.
3. Assess how well the compared text covers the reference text, \
and if any critical points are missing.
4. Assign a coverage score from 0.0 to 4.0 based on the provided criteria.

Important: The score must be a decimal number from 0.0 to 4.0. 4.0 is the maximum, \
do not exceed this value.
"""


INPUT_TEMPLATE: str = """
<REFERENCE_TEXT>
{reference}
</REFERENCE_TEXT>

<COMPARED_TEXT>
{compared}
</COMPARED_TEXT>
"""


@evaluator(name="text_coverage")
async def text_coverage_evaluator(
    compared: str,
    /,
    reference: str,
) -> EvaluationScore:
    if not compared:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference text was empty!",
        )

    score: CommonScoreModel = await generate_model(
        CommonScoreModel,
        instruction=INSTRUCTION,
        input=INPUT_TEMPLATE.format(
            reference=reference,
            compared=compared,
        ),
        examples=[
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Smartphones are versatile devices. They can make calls, send messages, "
                        "access the internet, take photos, and run various apps. "
                        "Many people use smartphones for work and entertainment. "
                        "However, excessive use can lead to addiction and sleep problems."
                    ),
                    compared=(
                        "Smartphones can make calls and send messages. They are popular devices."
                    ),
                ),
                CommonScoreModel(score=0.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Recycling helps protect the environment. It reduces waste in landfills, "
                        "conserves natural resources, and saves energy. Common recyclable items "
                        "include paper, plastic, glass, and metal. Many cities have recycling "
                        "programs, but individual participation is crucial for success."
                    ),
                    compared=(
                        "Recycling is good for the environment. "
                        "It reduces waste and saves resources. "
                        "People can recycle things like paper and plastic. "
                        "Many cities have recycling programs."
                    ),
                ),
                CommonScoreModel(score=2.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Regular exercise is important for health. It strengthens the heart, "
                        "builds muscle, and improves flexibility. Exercise can also reduce stress "
                        "and boost mood. Experts recommend at least 30 minutes of moderate "
                        "activity most days of the week. Walking, swimming, and cycling are "
                        "good options for many people."
                    ),
                    compared=(
                        "Regular exercise is crucial for maintaining good health. "
                        "It has many benefits, including strengthening the heart, "
                        "building muscle, and enhancing flexibility. Exercise also has "
                        "mental health benefits, such as reducing stress and improving mood. "
                        "Health experts advise doing at least 30 minutes of moderate exercise "
                        "on most days. Some popular and accessible forms of exercise "
                        "include walking, swimming, and cycling."
                    ),
                ),
                CommonScoreModel(score=4.0),
            ),
        ],
    )

    return score.normalized(divider=4)



from draive.sentencepiece.config import SentencePieceConfig
from draive.sentencepiece.tokenization import sentencepiece_tokenize_text

__all__ = [
    "sentencepiece_tokenize_text",
    "SentencePieceConfig",
]



import json
from asyncio import gather
from collections.abc import AsyncIterable, Sequence
from http import HTTPStatus
from itertools import chain
from typing import Any, Literal, Self, cast, final, overload

from httpx import AsyncClient, Response

from draive.mistral.config import MistralChatConfig, MistralEmbeddingConfig
from draive.mistral.errors import MistralException
from draive.mistral.models import (
    ChatCompletionResponse,
    ChatCompletionStreamResponse,
    ChatMessage,
    EmbeddingResponse,
)
from draive.parameters import DataModel
from draive.scope import ScopeDependency
from draive.utils import getenv_str, not_missing

__all__ = [
    "MistralClient",
]


@final
class MistralClient(ScopeDependency):
    @classmethod
    def prepare(cls) -> Self:
        return cls(
            endpoint=getenv_str("MISTRAL_ENDPOINT", "https://api.mistral.ai"),
            api_key=getenv_str("MISTRAL_API_KEY"),
            timeout=90,
        )

    def __init__(
        self,
        endpoint: str,
        api_key: str | None,
        timeout: float | None = None,
    ) -> None:
        self._client: AsyncClient = AsyncClient(
            base_url=endpoint,
            headers={
                "Authorization": f"Bearer {api_key}",
            },
            timeout=timeout,
        )

    @overload
    async def chat_completion(
        self,
        *,
        config: MistralChatConfig,
        messages: list[ChatMessage],
        tools: list[dict[str, object]] | None = None,
        tool_choice: Literal["auto", "any", "none"] = "auto",
        stream: Literal[True],
    ) -> AsyncIterable[ChatCompletionStreamResponse]: ...

    @overload
    async def chat_completion(
        self,
        *,
        config: MistralChatConfig,
        messages: list[ChatMessage],
        tools: list[dict[str, object]] | None = None,
        tool_choice: Literal["auto", "any", "none"] = "auto",
    ) -> ChatCompletionResponse: ...

    async def chat_completion(
        self,
        *,
        config: MistralChatConfig,
        messages: list[ChatMessage],
        tools: list[dict[str, object]] | None = None,
        tool_choice: Literal["auto", "any", "none"] = "auto",
        stream: bool = False,
    ) -> AsyncIterable[ChatCompletionStreamResponse] | ChatCompletionResponse:
        if stream:
            raise NotImplementedError("Mistral streaming is not supported yet")

        else:
            return await self._create_chat_completion(
                messages=messages,
                model=config.model,
                temperature=config.temperature,
                top_p=config.top_p if not_missing(config.top_p) else None,
                max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
                response_format=cast(dict[str, str], config.response_format),
                seed=config.seed if not_missing(config.seed) else None,
                tools=tools,
                tool_choice=tool_choice if tools else None,
            )

    async def embedding(
        self,
        config: MistralEmbeddingConfig,
        inputs: Sequence[str],
    ) -> list[list[float]]:
        return list(
            chain(
                *await gather(
                    *[
                        self._create_text_embedding(
                            model=config.model,
                            texts=inputs[index : index + config.batch_size],
                        )
                        for index in range(0, len(inputs), config.batch_size)
                    ]
                )
            )
        )

    async def _create_chat_completion(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_p: float | None,
        seed: int | None,
        max_tokens: int | None,
        response_format: dict[str, str] | None,
        messages: list[ChatMessage],
        tools: list[dict[str, object]] | None,
        tool_choice: str | None,
    ) -> ChatCompletionResponse:
        request_body: dict[str, Any] = {
            "model": model,
            "temperature": temperature,
            "messages": messages,
        }

        if tools:
            request_body["tools"] = tools
        if tool_choice is not None:
            request_body["tool_choice"] = tool_choice
        elif tools:
            request_body["tool_choice"] = "auto"
        if max_tokens:
            request_body["max_tokens"] = max_tokens
        if top_p is not None:
            request_body["top_p"] = top_p
        if seed is not None:
            request_body["random_seed"] = seed
        if response_format is not None:
            request_body["response_format"] = response_format

        return await self._request(
            model=ChatCompletionResponse,
            method="POST",
            url="v1/chat/completions",
            body=request_body,
        )

    async def _create_text_embedding(
        self,
        model: str,
        texts: Sequence[str],
    ) -> list[list[float]]:
        response: EmbeddingResponse = await self._request(
            model=EmbeddingResponse,
            method="POST",
            url="v1/embeddings",
            body={
                "model": model,
                "input": texts,
            },
        )
        return [element.embedding for element in response.data]

    async def dispose(self) -> None:
        await self._client.aclose()

    async def _request[Requested: DataModel](  # noqa: PLR0913
        self,
        model: type[Requested],
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> Requested:
        request_headers: dict[str, str]
        if headers:
            request_headers = headers

        else:
            request_headers = {
                "Accept": "application/json",
            }

        body_content: str | None
        match body:
            case None:
                body_content = None

            case body_model if isinstance(body_model, DataModel):
                body_content = body_model.as_json()

            case values:
                body_content = json.dumps(values)

        if body_content:
            request_headers["Content-Type"] = "application/json"

        response: Response
        try:
            response = await self._client.request(
                method=method,
                url=url,
                headers=request_headers,
                params=query,
                content=body_content,
                follow_redirects=follow_redirects or False,
                timeout=timeout,
            )

        except Exception as exc:
            raise MistralException("Network request failed") from exc

        status: HTTPStatus = HTTPStatus(value=response.status_code)
        if status.is_success:
            try:
                return model.from_json(await response.aread())

            except Exception as exc:
                raise MistralException("Failed to decode Mistral response %s", response) from exc

        elif status.is_client_error:
            error_body: bytes = await response.aread()
            raise MistralException(
                "Mistral request error: %s %s",
                status,
                error_body.decode("utf-8"),
            )

        else:
            raise MistralException("Network request failed %s", response)
